from collections import Counter
import datetime
import json
import logging
from pathlib import Path
import os
import shutil
import time
from unittest.mock import MagicMock, AsyncMock

import htcondor2 as htcondor
import pytest

from iceprod.core.config import Job, Task
from iceprod.core.exe import Data, Transfer
import iceprod.server.config
import iceprod.server.grid
from iceprod.server.util import datetime2str
import iceprod.server.plugins.condor
from iceprod.server.plugins.condor import CondorJob, CondorJobId, JobStatus

htcondor.enable_debug()


@pytest.fixture
def schedd(monkeypatch):
    mock = MagicMock()
    monkeypatch.setattr(htcondor, 'Schedd', mock)
    yield mock


@pytest.fixture
def credd(monkeypatch):
    mock = MagicMock()
    monkeypatch.setattr(htcondor, 'Credd', mock)
    yield mock


@pytest.fixture
def set_time(monkeypatch):
    now = datetime.datetime(2024, 10, 10, 10, 50, 0, 0, datetime.UTC)
    mock = MagicMock()
    mock.now = MagicMock(return_value=now)
    monkeypatch.setattr(iceprod.server.plugins.condor, 'datetime', mock)
    tnow = time.mktime(now.utctimetuple())
    tmock = MagicMock(return_value=tnow)
    monkeypatch.setattr(time, 'time', tmock)
    yield now


def test_grid_init(schedd):
    override = ['queue.type=condor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)


def test_CondorJobId():
    j1 = CondorJobId(cluster_id=0, proc_id=0)
    assert str(j1) == '0.0'


def test_CondorSubmit_init(schedd):
    override = ['queue.type=condor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)
    submit_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['submit_dir'])))
    cred_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['credentials_dir'])))

    sub = iceprod.server.plugins.condor.CondorSubmit(cfg=cfg, submit_dir=submit_dir, credentials_dir=cred_dir)
    assert set(sub.transfer_plugins.keys()) == {'gsiftp', 'iceprod-plugin'}


@pytest.mark.parametrize("os_arch,container", [
    ('RHEL_6_x86_64', 'el6'),
    ('RHEL_7_x86_64', 'el7'),
    ('RHEL_8_x86_64', 'el8'),
    ('RHEL_9_x86_64', 'el9'),
    ('RHEL_9_x86_64_v2', 'el9'),
])
def test_CondorSubmit_condor_os_container(os_arch, container):
    ret = iceprod.server.plugins.condor.CondorSubmit.condor_os_container(os_arch)
    assert container in ret


def test_CondorSubmit_condor_resource_reqs():
    task = Task(
        dataset=MagicMock(),
        job=MagicMock(),
        task_id='task',
        task_index=0,
        name='generate',
        depends=[],
        requirements={'cpu': 1},
        status='queued',
        site='site',
        stats={},
    )

    ret = iceprod.server.plugins.condor.CondorSubmit.condor_resource_reqs(task)
    assert ret['request_cpus'] == 1
    assert 'request_memory' not in ret
    assert 'request_gpus' not in ret
    assert 'requirements' not in ret

    task.requirements = {
        'gpu': 1,
    }
    ret = iceprod.server.plugins.condor.CondorSubmit.condor_resource_reqs(task)
    assert ret['request_gpus'] == 1


def test_CondorSubmit_condor_infiles(schedd):
    override = ['queue.type=condor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)
    submit_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['submit_dir'])))
    cred_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['credentials_dir'])))

    sub = iceprod.server.plugins.condor.CondorSubmit(cfg=cfg, submit_dir=submit_dir, credentials_dir=cred_dir)

    infiles = [
        Data(url='http://foo.test/foo', local='foo', transfer=Transfer.TRUE)
    ]

    ret = sub.condor_infiles(infiles, {})
    assert ret['transfer_input_files'] == ['iceprod-plugin://true-http://foo.test/foo']
    assert 'PreCmd' not in ret


def test_CondorSubmit_condor_infiles_maybe(schedd):
    override = ['queue.type=condor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)
    submit_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['submit_dir'])))
    cred_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['credentials_dir'])))

    sub = iceprod.server.plugins.condor.CondorSubmit(cfg=cfg, submit_dir=submit_dir, credentials_dir=cred_dir)

    infiles = [
        Data(url='http://foo.test/foo', local='foo', transfer=Transfer.MAYBE)
    ]

    ret = sub.condor_infiles(infiles, {})
    assert ret['transfer_input_files'] == ['iceprod-plugin://maybe-http://foo.test/foo']
    assert 'PreCmd' not in ret


def test_CondorSubmit_condor_infiles_gsiftp(schedd):
    override = ['queue.type=condor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)
    submit_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['submit_dir'])))
    cred_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['credentials_dir'])))

    sub = iceprod.server.plugins.condor.CondorSubmit(cfg=cfg, submit_dir=submit_dir, credentials_dir=cred_dir)

    infiles = [
        Data(url='gsiftp://foo.test/foo', local='foo', transfer=Transfer.TRUE)
    ]

    with pytest.raises(RuntimeError, match='x509 proxy'):
        sub.condor_infiles(infiles, {})

    cfg['queue']['x509proxy'] = '/tmp/x509'

    ret = sub.condor_infiles(infiles, {})
    assert ret['transfer_input_files'] == ['/tmp/x509', 'iceprod-plugin://true-gsiftp://foo.test/foo']
    assert 'PreCmd' not in ret


# skip this because we're using the iceprod transfer plugin
@pytest.mark.skip
def test_CondorSubmit_condor_precmd(schedd, i3prod_path):
    logging.info('cfgfile: %r', os.path.exists(os.path.expandvars('$I3PROD/etc/iceprod_config.json')))
    override = ['queue.type=condor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)
    submit_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['submit_dir'])))
    cred_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['credentials_dir'])))

    sub = iceprod.server.plugins.condor.CondorSubmit(cfg=cfg, submit_dir=submit_dir, credentials_dir=cred_dir)

    infiles = [
        Data(url='http://foo.test/foo', local='bar', transfer=Transfer.TRUE)
    ]

    logging.info('cfg: %r', cfg)

    ret = sub.condor_infiles(infiles, {})
    logging.info('ret: %r', ret)
    assert ret['transfer_input_files'][0] == 'http://foo.test/foo'
    assert 'PreCmd' in ret
    assert 'PreArguments' in ret
    assert 'bar' in ret['PreArguments']


def test_CondorSubmit_condor_outfiles(schedd):
    override = ['queue.type=condor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)
    submit_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['submit_dir'])))
    cred_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['credentials_dir'])))

    sub = iceprod.server.plugins.condor.CondorSubmit(cfg=cfg, submit_dir=submit_dir, credentials_dir=cred_dir)

    outfiles = [
        Data(url='http://foo.test/foo', local='bar', transfer=Transfer.TRUE)
    ]

    ret = sub.condor_outfiles(outfiles, {})
    assert ret['transfer_output_files'] == ['bar']
    assert ret['transfer_output_remaps'] == 'bar = http://foo.test/foo'


def test_CondorSubmit_condor_outfiles_maybe(schedd):
    override = ['queue.type=condor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)
    submit_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['submit_dir'])))
    cred_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['credentials_dir'])))

    sub = iceprod.server.plugins.condor.CondorSubmit(cfg=cfg, submit_dir=submit_dir, credentials_dir=cred_dir)

    outfiles = [
        Data(url='http://foo.test/foo', local='bar', transfer=Transfer.MAYBE)
    ]

    ret = sub.condor_outfiles(outfiles, {})
    assert ret['transfer_output_files'] == ['bar']
    assert ret['transfer_output_remaps'] == 'bar = iceprod-plugin://maybe-http://foo.test/foo'


def test_CondorSubmit_condor_oauth_transform(schedd):
    override = ['queue.type=condor', 'oauth_services.token://=ttt', 'oauth_services.token2=t2']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)
    submit_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['submit_dir'])))
    cred_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['credentials_dir'])))

    sub = iceprod.server.plugins.condor.CondorSubmit(cfg=cfg, submit_dir=submit_dir, credentials_dir=cred_dir)

    handle = 'handle'
    scopes = {
        'token://': 'storage.read:/ storage.write:/data'
    }
    services = sub.condor_oauth_url_transform(handle, scopes)
    assert services == {'token://': 'ttt.handle+token://'}

    scopes = {
        'token://': 'storage.read:/ storage.write:/data',
        'token2': 'storage.read:/'
    }
    services = sub.condor_oauth_url_transform(handle, scopes)
    assert services == {'token://': 'ttt.handle+token://', 'token2': 't2.handle+token2'}

    scopes = {
        'token3': 'storage.read:/'
    }
    with pytest.raises(RuntimeError, match='unknown token scope url prefix'):
        sub.condor_oauth_url_transform(handle, scopes)


def test_CondorSubmit_condor_oauth_scratch(schedd):
    override = ['queue.type=condor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)
    cfg['queue']['x509proxy'] = '/tmp/x509'
    cfg['oauth_services'] = {
        'token://foo.bar': 'token'
    }
    submit_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['submit_dir'])))
    cred_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['credentials_dir'])))

    sub = iceprod.server.plugins.condor.CondorSubmit(cfg=cfg, submit_dir=submit_dir, credentials_dir=cred_dir)

    cred_dir.mkdir(parents=True)
    cred_path = cred_dir / 'scratch'
    exp = time.time() + 100
    cred = {"type":"oauth","refresh_token":"refresh","access_token":"access","expiration":exp}
    with open(cred_path, 'w') as f:
        json.dump([cred], f)

    dataset = MagicMock()
    dataset.dataset_id = 'dataset'
    dataset.dataset_num = 0
    dataset.config = {
        'options': {
            'site_temp': 'token://foo.bar/scratch',
        },
        'steering': {
            'parameters': {},
        },
        'tasks': [{
            'name': 'generate',
            'batchsys': {},
            'requirements': {},
            'token_scopes': {},
            'trays': [{
                'iterations': 1,
                'modules': [{
                    'name': 'foo',
                    'src': 'foo.py',
                    'args': '',
                    'env_shell': None,
                    'env_clear': True,
                    'configs': None,
                }]
            }],
            'data': [{
                'movement': 'output',
                'local': 'foo.tgz',
                'remote': '',
                'type': 'job_temp',
                'transfer': True,
            }]
        }]
    }
    dataset.fill_defaults()
    dataset.validate()
    logging.info('config: %r', dataset.config)
    job = Job(
        dataset=dataset,
        job_id='job',
        job_index=1,
        status='processing',
    )
    task = Task(
        dataset=dataset,
        job=job,
        task_id='task',
        task_index=0,
        name='generate',
        depends=[],
        requirements={'cpu': 1},
        status='queued',
        site='site',
        stats={},
    )

    ret = sub.condor_oauth_scratch(task)
    assert ret
    prefix, ret_cred = ret
    assert prefix == 'token://foo.bar'
    assert ret_cred == cred


def test_CondorSubmit_add_oauth_tokens(schedd, credd):
    override = ['queue.type=condor', 'queue.site_temp=http://foo.bar']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)
    cfg['queue']['x509proxy'] = '/tmp/x509'
    cfg['oauth_services'] = {
        'token://foo.bar': 'token'
    }
    submit_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['submit_dir'])))
    cred_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['credentials_dir'])))

    sub = iceprod.server.plugins.condor.CondorSubmit(cfg=cfg, submit_dir=submit_dir, credentials_dir=cred_dir)

    transforms = {
        'token://foo.bar': 'token.datasettaskname+token://foo.bar'
    }
    tokens = [
        {
            'url': 'http://issuer.bar',
            'type': 'oauth',
            'transfer_prefix': 'token://foo.bar',
            'access_token': 'access',
            'refresh_token': 'refresh',
            'scope': 'storage.modify:/baz',
            'expiration': time.time() + 10.4,
        }
    ]

    add_cred = credd.return_value.add_user_service_cred = MagicMock()

    sub.add_oauth_tokens(transforms, tokens)

    logging.info('check_call_args: %r', add_cred.call_args_list)
    assert add_cred.call_count == 2
    assert add_cred.call_args_list[0].kwargs['refresh'] == False
    assert json.loads(add_cred.call_args_list[0].kwargs['credential']) == {
        "access_token": "access",
        "token_type": "bearer",
        "expires_in": 10,
        "expires_at": tokens[0]["expiration"],
        "scope": tokens[0]["scope"].split(),
    }
    assert add_cred.call_args_list[1].kwargs['refresh'] == True
    assert json.loads(add_cred.call_args_list[1].kwargs['credential']) == {
        "refresh_token": "refresh",
        "scopes": tokens[0]["scope"],
    }


async def test_CondorSubmit_submit(schedd):
    override = ['queue.type=condor', 'queue.site_temp=pelican://foo.bar/scratch']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)
    cfg['oauth_services'] = {
        'osdf:///': 'token',
        'pelican://foo.bar': 'pelican',
    }
    submit_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['submit_dir'])))
    cred_dir = Path(os.path.expanduser(os.path.expandvars(cfg['queue']['credentials_dir'])))

    cred_dir.mkdir(parents=True)
    cred_path = cred_dir / 'scratch'
    exp = time.time() + 100
    cred = {"type":"oauth","refresh_token":"refresh","access_token":"access","expiration":exp}
    with open(cred_path, 'w') as f:
        json.dump([cred], f)

    sub = iceprod.server.plugins.condor.CondorSubmit(cfg=cfg, submit_dir=submit_dir, credentials_dir=cred_dir)
    sub.condor_schedd.submit = MagicMock()
    sub._restart_schedd = MagicMock()
    sub.add_oauth_tokens = MagicMock()

    dataset = MagicMock()
    dataset.dataset_id = 'dataset'
    dataset.dataset_num = 0
    dataset.config = {
        'options': {
            'site_temp': 'pelican://foo.bar/scratch',
        },
        'steering': {
            'parameters': {},
        },
        'tasks': [{
            'name': 'generate',
            'batchsys': {},
            'requirements': {},
            'token_scopes': {
                'osdf:///': 'storage.modify:/baz',
            },
            'trays': [{
                'iterations': 1,
                'modules': [{
                    'name': 'foo',
                    'src': 'foo.py',
                    'args': '',
                    'env_shell': None,
                    'env_clear': True,
                    'configs': None,
                }]
            }],
            'data': [{
                'movement': 'output',
                'local': 'foo.tgz',
                'remote': '',
                'type': 'job_temp',
                'transfer': True,
            }, {
                'movement': 'input',
                'type': 'permanent',
                'local': '',
                'remote': 'gsiftp://foo.bar/baz',
                'transfer': 'maybe',
            }, {
                'movement': 'output',
                'type': 'permanent',
                'local': '',
                'remote': 'osdf:///baz',
                'transfer': 'maybe',
            }]
        }]
    }
    job = Job(
        dataset=dataset,
        job_id='job',
        job_index=1,
        status='processing',
    )
    tokens = [
        {
            'url': 'http://issuer.bar',
            'transfer_prefix': 'osdf:///',
            'access_token': 'access',
            'refresh_token': 'refresh',
        }
    ]
    task = Task(
        dataset=dataset,
        job=job,
        task_id='task',
        task_index=0,
        name='generate',
        depends=[],
        requirements={'cpu': 1},
        status='queued',
        site='site',
        stats={},
        oauth_tokens=tokens,
    )

    jobs = [
        task
    ]

    jel = submit_dir / 'today' / 'condor.log'
    await sub.submit(jobs, jel=jel)

    assert (submit_dir / 'today' / 'task').is_dir()

    assert sub.condor_schedd.submit.call_count == 1
    itemdata = list(sub.condor_schedd.submit.call_args.kwargs['itemdata'])[0]
    logging.info('itemdata: %r', itemdata)
    assert itemdata['infiles'].strip('"') == ''
    assert itemdata['outfiles'].strip('"') == ''
    assert itemdata['outremaps'].strip('"') == ''

    assert sub.add_oauth_tokens.call_count == 2
    assert sub.add_oauth_tokens.call_args_list[1].args == ({'osdf:///': 'token.datasetgenerate+osdf:///'}, tokens)
    assert sub.add_oauth_tokens.call_args_list[0].args == ({'pelican://foo.bar/scratch': 'pelican.scratch+pelican://foo.bar/scratch'}, [cred])

    submitfile = sub.condor_schedd.submit.call_args.args[0]
    logging.info('submitfile: %s', submitfile)
    assert submitfile['My.OAuthServicesNeeded'] == '"token*datasetgenerate pelican*scratch"'

    exe = open(submit_dir / 'today' / 'task' / 'task_runner.sh').read()
    logging.info('exe: %s', exe)
    assert 'object put foo.tgz pelican://foo.bar/scratch/0/1/foo.tgz' in exe
    assert 'if [ -f baz ]' in exe
    assert 'object put baz osdf:///baz' in exe

async def test_Grid_save_load_timestamp(schedd, i3prod_path):
    override = ['queue.type=htcondor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    g.last_event_timestamp = 12345.
    g.save_timestamp()

    g.last_event_timestamp = 0
    g.load_timestamp()
    assert g.last_event_timestamp == 12345.


async def test_Grid_run(schedd, i3prod_path):
    override = ['queue.type=htcondor', 'queue.submit_interval=0', 'queue.check_time=0']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    g.submit = AsyncMock()
    g.wait = AsyncMock()
    g.check = AsyncMock()

    await g.run(forever=False)
    assert g.submit.call_count == 1
    assert g.wait.call_count == 2
    assert g.check.call_count == 2


async def test_Grid_run_error(schedd, i3prod_path):
    override = ['queue.type=htcondor', 'queue.submit_interval=0', 'queue.check_time=0']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    g.submit = AsyncMock(side_effect=RuntimeError())
    g.wait = AsyncMock(side_effect=RuntimeError())
    g.check = AsyncMock(side_effect=RuntimeError())

    await g.run(forever=False)
    assert g.submit.call_count == 1
    assert g.wait.call_count == 2
    assert g.check.call_count == 2


def test_Grid_queue_dataset_status(schedd, i3prod_path):
    override = ['queue.type=htcondor', 'queue.submit_interval=0', 'queue.check_time=0']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    g.submit = AsyncMock(side_effect=RuntimeError())
    g.wait = AsyncMock(side_effect=RuntimeError())
    g.check = AsyncMock(side_effect=RuntimeError())

    ret = g.queue_dataset_status()
    assert not ret

    g.jobs[CondorJobId(cluster_id=1, proc_id=0)] = CondorJob(status=JobStatus.IDLE, dataset_id='d1')
    ret = g.queue_dataset_status()
    assert ret == {iceprod.server.grid.GridStatus.QUEUED: {'d1': 1}}

    g.jobs[CondorJobId(cluster_id=2, proc_id=0)] = CondorJob(status=JobStatus.RUNNING, dataset_id='d1')
    g.jobs[CondorJobId(cluster_id=2, proc_id=1)] = CondorJob(status=JobStatus.RUNNING, dataset_id='d1')
    ret = g.queue_dataset_status()
    assert ret == {iceprod.server.grid.GridStatus.QUEUED: {'d1': 1},
                   iceprod.server.grid.GridStatus.PROCESSING: {'d1': 2}}

    g.jobs[CondorJobId(cluster_id=3, proc_id=0)] = CondorJob(status=JobStatus.COMPLETED, dataset_id='d1')
    ret = g.queue_dataset_status()
    assert ret == {iceprod.server.grid.GridStatus.QUEUED: {'d1': 1},
                   iceprod.server.grid.GridStatus.PROCESSING: {'d1': 2}}

    g.jobs[CondorJobId(cluster_id=4, proc_id=0)] = CondorJob(status=JobStatus.RUNNING, dataset_id='d2')
    ret = g.queue_dataset_status()
    assert ret == {iceprod.server.grid.GridStatus.QUEUED: {'d1': 1},
                   iceprod.server.grid.GridStatus.PROCESSING: {'d1': 2, 'd2': 1}}


async def test_Grid_submit(schedd, i3prod_path):
    override = ['queue.type=htcondor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)
    g.get_scratch_credentials = AsyncMock()

    g.get_queue_num = MagicMock(return_value=5)
    tasks = [MagicMock(), MagicMock()]
    g.get_tasks_to_queue = AsyncMock(return_value=tasks)
    jel_path = i3prod_path/'today'/'jel.log'
    jel_path.parent.mkdir()
    g.get_current_JEL = MagicMock(return_value=jel_path)
    g.submitter.submit = AsyncMock(return_value={})
    g.task_reset = AsyncMock()

    await g.submit()
    assert g.get_queue_num.call_count == 1
    assert g.get_tasks_to_queue.call_count == 1
    assert g.get_current_JEL.call_count == 1
    assert g.submitter.submit.call_count == 2
    assert g.task_reset.call_count == 0


async def test_Grid_submit_error(schedd, i3prod_path):
    override = ['queue.type=htcondor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)
    g.get_scratch_credentials = AsyncMock()

    g.get_queue_num = MagicMock(return_value=5)
    tasks = [MagicMock(), MagicMock()]
    g.get_tasks_to_queue = AsyncMock(return_value=tasks)
    jel_path = i3prod_path/'today'/'jel.log'
    jel_path.parent.mkdir()
    g.get_current_JEL = MagicMock(return_value=jel_path)
    g.submitter.submit = AsyncMock(side_effect=RuntimeError())
    g.task_reset = AsyncMock()

    await g.submit()
    assert g.get_queue_num.call_count == 1
    assert g.get_tasks_to_queue.call_count == 1
    assert g.get_current_JEL.call_count == 1
    assert g.submitter.submit.call_count == 2
    assert g.task_reset.call_count == 2


async def test_Grid_get_queue_num(schedd, i3prod_path):
    override = ['queue.type=htcondor', 'queue.max_total_tasks_on_queue=10', 'queue.max_idle_tasks_on_queue=5', 'queue.max_tasks_per_submit=3']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    n = g.get_queue_num()
    assert n == 3

    g.jobs[CondorJobId(cluster_id=1, proc_id=0)] = CondorJob(status=JobStatus.IDLE)
    g.jobs[CondorJobId(cluster_id=2, proc_id=0)] = CondorJob(status=JobStatus.IDLE)
    g.jobs[CondorJobId(cluster_id=3, proc_id=0)] = CondorJob(status=JobStatus.IDLE)
    n = g.get_queue_num()
    assert n == 2

    g.jobs[CondorJobId(cluster_id=4, proc_id=0)] = CondorJob(status=JobStatus.IDLE)
    g.jobs[CondorJobId(cluster_id=5, proc_id=0)] = CondorJob(status=JobStatus.IDLE)
    n = g.get_queue_num()
    assert n == 0

    g.jobs[CondorJobId(cluster_id=1, proc_id=0)] = CondorJob(status=JobStatus.RUNNING)
    g.jobs[CondorJobId(cluster_id=2, proc_id=0)] = CondorJob(status=JobStatus.RUNNING)
    g.jobs[CondorJobId(cluster_id=3, proc_id=0)] = CondorJob(status=JobStatus.RUNNING)
    g.jobs[CondorJobId(cluster_id=4, proc_id=0)] = CondorJob(status=JobStatus.RUNNING)
    g.jobs[CondorJobId(cluster_id=5, proc_id=0)] = CondorJob(status=JobStatus.RUNNING)
    g.jobs[CondorJobId(cluster_id=6, proc_id=0)] = CondorJob(status=JobStatus.RUNNING)
    g.jobs[CondorJobId(cluster_id=7, proc_id=0)] = CondorJob(status=JobStatus.RUNNING)
    n = g.get_queue_num()
    assert n == 3

    g.jobs[CondorJobId(cluster_id=8, proc_id=0)] = CondorJob(status=JobStatus.IDLE)
    g.jobs[CondorJobId(cluster_id=9, proc_id=0)] = CondorJob(status=JobStatus.IDLE)
    n = g.get_queue_num()
    assert n == 1

    g.jobs[CondorJobId(cluster_id=10, proc_id=0)] = CondorJob(status=JobStatus.IDLE)
    n = g.get_queue_num()
    assert n == 0

    g.jobs[CondorJobId(cluster_id=11, proc_id=0)] = CondorJob(status=JobStatus.IDLE)
    n = g.get_queue_num()
    assert n == 0


async def test_Grid_get_current_JEL(schedd, i3prod_path, set_time):
    override = ['queue.type=htcondor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    ret = g.get_current_JEL()
    assert set_time.strftime('%Y-%m-%d') in str(ret) 

    assert len(g.jels) == 1


async def test_Grid_wait_no_events(schedd, i3prod_path, set_time):
    override = ['queue.type=htcondor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    await g.wait(timeout=0)


async def test_Grid_wait_JEL(schedd, i3prod_path, set_time):
    override = ['queue.type=htcondor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    g.task_idle = AsyncMock()
    g.task_processing = AsyncMock()
    g.task_reset = AsyncMock()
    g.finish = AsyncMock()

    jel_path = g.get_current_JEL()
    TEST_JEL = Path(__file__).parent / 'condor_test_logfile'
    shutil.copy(TEST_JEL, jel_path)

    # 0 = success, transfer output
    # 1 = success, input and output
    # 2 = success, chirp to ulog
    # 3 = failure
    # 4 = hold, memory exceeded
    # 5 = success, job evicted
    # 6 = condor_rm

    await g.wait(timeout=0)

    #assert len(g.jobs) == 7

    assert g.jobs[CondorJobId(cluster_id=110828038, proc_id=0)].dataset_id == '4ksd8'
    assert g.jobs[CondorJobId(cluster_id=110828038, proc_id=0)].task_id == 'lnk3f'
    assert g.jobs[CondorJobId(cluster_id=110828038, proc_id=0)].submit_dir == Path('/scratch/dschultz')
    #assert g.jobs[CondorJobId(cluster_id=110828038, proc_id=0)].status == JobStatus.COMPLETED
    
    #assert g.jobs[CondorJobId(cluster_id=110828038, proc_id=1)].status == JobStatus.COMPLETED
    #assert g.jobs[CondorJobId(cluster_id=110828038, proc_id=2)].status == JobStatus.COMPLETED
    #assert g.jobs[CondorJobId(cluster_id=110828038, proc_id=3)].status == JobStatus.FAILED
    #assert g.jobs[CondorJobId(cluster_id=110828038, proc_id=4)].status == JobStatus.FAILED
    #assert g.jobs[CondorJobId(cluster_id=110828038, proc_id=5)].status == JobStatus.COMPLETED
    #assert g.jobs[CondorJobId(cluster_id=110828038, proc_id=6)].status == JobStatus.FAILED

    assert g.task_idle.call_count == 1
    assert g.task_processing.call_count == 7
    assert g.task_reset.call_count == 0
    #assert g.finish.call_count == 6


async def test_Grid_wait_JEL_finish(schedd, i3prod_path, set_time):
    override = ['queue.type=htcondor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    g.task_idle = AsyncMock()
    g.task_processing = AsyncMock()
    g.task_reset = AsyncMock()
    g.task_failure = AsyncMock()
    g.task_success = AsyncMock()

    jel_path = g.get_current_JEL()
    TEST_JEL = Path(__file__).parent / 'condor_test_logfile'
    shutil.copy(TEST_JEL, jel_path)

    await g.wait(timeout=0)

    #assert len(g.jobs) == 1
    assert CondorJobId(cluster_id=110828038, proc_id=4) in g.jobs

    assert g.task_idle.call_count == 1
    assert g.task_processing.call_count == 7
    assert g.task_reset.call_count == 0
    #assert g.task_failure.call_count == 2
    #assert g.task_success.call_count == 4


async def test_Grid_wait_JEL_exception(schedd, i3prod_path, set_time):
    override = ['queue.type=htcondor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    g.task_idle = AsyncMock(side_effect=RuntimeError)
    g.task_processing = AsyncMock(side_effect=RuntimeError)
    g.task_reset = AsyncMock(side_effect=RuntimeError)
    g.task_failure = AsyncMock(side_effect=RuntimeError)
    g.task_success = AsyncMock(side_effect=RuntimeError)

    jel_path = g.get_current_JEL()
    TEST_JEL = Path(__file__).parent / 'condor_test_logfile'
    shutil.copy(TEST_JEL, jel_path)

    await g.wait(timeout=0)

    #assert len(g.jobs) == 1
    assert CondorJobId(cluster_id=110828038, proc_id=4) in g.jobs

    assert g.task_idle.call_count == 1
    assert g.task_processing.call_count == 7
    assert g.task_reset.call_count == 0
    #assert g.task_failure.call_count == 2
    #assert g.task_success.call_count == 4


async def test_Grid_wait_JEL_reprocess(schedd, i3prod_path, set_time):
    override = ['queue.type=htcondor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    g.task_idle = AsyncMock()
    g.task_processing = AsyncMock()
    g.task_reset = AsyncMock()
    g.finish = AsyncMock()

    g.last_event_timestamp = 1710261004

    jel_path = g.get_current_JEL()
    TEST_JEL = Path(__file__).parent / 'condor_test_logfile'
    shutil.copy(TEST_JEL, jel_path)

    await g.wait(timeout=0)

    assert int(g.last_event_timestamp) == 1710261004
    assert len(g.jobs) == 0


async def test_Grid_check_empty(schedd, i3prod_path, set_time):
    override = ['queue.type=htcondor']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    g.submitter.get_jobs = MagicMock(return_value={})
    g.submitter.get_history = MagicMock(return_value={})
    g.submitter.remove = MagicMock()
    g.get_tasks_on_queue = AsyncMock(return_value=[])

    await g.check()

    dirs = {x.name: list(x.iterdir()) for x in g.submit_dir.glob('[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]T[0-9][0-9]')}
    assert dirs == {}


async def test_Grid_check_delete_day(schedd, i3prod_path, set_time):
    override = ['queue.type=htcondor', 'queue.max_task_queued_time=10', 'queue.max_task_processing_time=10', 'queue.suspend_submit_dir_time=10']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    g.submitter.get_jobs = MagicMock(return_value={})
    g.submitter.get_history = MagicMock(return_value={})
    g.submitter.remove = MagicMock()
    g.get_tasks_on_queue = AsyncMock(return_value=[])

    jel = g.get_current_JEL()
    p = jel.parent
    t = time.mktime(set_time.utctimetuple())
    
    assert g.jels != {}

    await g.check()

    dirs = {x.name: [x for x in x.iterdir() if x.is_dir()] for x in g.submit_dir.glob('[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]T[0-9][0-9]')}
    assert dirs == {}
    assert g.jels == {}


async def test_Grid_check_old_delete(schedd, i3prod_path, set_time):
    override = ['queue.type=htcondor', 'queue.max_task_queued_time=10', 'queue.max_task_processing_time=10', 'queue.suspend_submit_dir_time=10']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    g.submitter.get_jobs = MagicMock(return_value={})
    g.submitter.get_history = MagicMock(return_value={})
    g.submitter.remove = MagicMock()
    g.get_tasks_on_queue = AsyncMock(return_value=[])

    jel = g.get_current_JEL()
    daydir = jel.parent
    p = daydir / 'olddir'
    p.mkdir()
    t = time.mktime(set_time.utctimetuple()) - 35  # must be older than all times added together
    os.utime(p, (t, t))
    logging.info('set time to %d', t)

    await g.check()

    dirs = {x.name: [x for x in x.iterdir() if x.is_dir()] for x in g.submit_dir.glob('[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]T[0-9][0-9]')}
    assert dirs == {daydir.name: []}


async def test_Grid_check_oldjob_remove(schedd, i3prod_path, set_time):
    override = ['queue.type=htcondor', 'queue.max_task_queued_time=10', 'queue.max_task_processing_time=10', 'queue.suspend_submit_dir_time=10']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    jobs = {}
    g.submitter.get_jobs = MagicMock(return_value=jobs)
    g.submitter.get_history = MagicMock(return_value={})
    g.submitter.remove = MagicMock()
    g.get_tasks_on_queue = AsyncMock(return_value=[])

    jel = g.get_current_JEL()
    daydir = jel.parent
    p = daydir / 'olddir'
    p.mkdir()
    t = time.mktime(set_time.utctimetuple()) - 25  # must be older than queued + processing time
    os.utime(p, (t, t))
    logging.info('set time to %d', t)

    jobs[CondorJobId(cluster_id=1, proc_id=0)] = CondorJob(status=JobStatus.IDLE, submit_dir=p, task_id=p.name)

    await g.check()

    dirs = {x.name: [x for x in x.iterdir() if x.is_dir()] for x in g.submit_dir.glob('[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]T[0-9][0-9]')}
    assert dirs == {daydir.name: [p]}
    assert g.submitter.remove.call_count == 1


async def test_Grid_check_oldjob_delete(schedd, i3prod_path, set_time):
    override = ['queue.type=htcondor', 'queue.max_task_queued_time=10', 'queue.max_task_processing_time=10', 'queue.suspend_submit_dir_time=10']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    jobs = {}
    g.submitter.get_jobs = MagicMock(return_value=jobs)
    g.submitter.get_history = MagicMock(return_value={})
    g.submitter.remove = MagicMock()
    g.get_tasks_on_queue = AsyncMock(return_value=[])

    jel = g.get_current_JEL()
    daydir = jel.parent
    p = daydir / 'olddir'
    p.mkdir()
    t = time.mktime(set_time.utctimetuple()) - 15  # must be older than suspend time
    os.utime(p, (t, t))
    logging.info('set time to %d', t)

    await g.check()

    dirs = {x.name: [x for x in x.iterdir() if x.is_dir()] for x in g.submit_dir.glob('[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]T[0-9][0-9]')}
    assert dirs == {daydir.name: []}
    assert g.submitter.remove.call_count == 0


@pytest.mark.parametrize('jel_jobs,queue_jobs,hist_jobs,remove_calls,finish_calls', [
    ({(1,0): JobStatus.IDLE},
     {(1,0): JobStatus.IDLE},
     {},
     0, 0
    ),
    ({(1,0): JobStatus.IDLE},
     {},
     {(1,0): JobStatus.COMPLETED},
     0, 1
    ),
    ({(1,0): JobStatus.IDLE},
     {(1,0): JobStatus.COMPLETED},
     {(1,0): JobStatus.COMPLETED},
     0, 1
    ),
    ({(1,0): JobStatus.IDLE},
     {(1,0): JobStatus.FAILED},
     {},
     1, 0
    ),
    ({},
     {(1,0): JobStatus.IDLE},
     {},
     0, 0
    ),
    ({},
     {(1,0): JobStatus.RUNNING},
     {},
     0, 0
    ),
    ({},
     {(1,0): JobStatus.FAILED},
     {},
     1, 0
    ),
    ({},
     {(1,0): JobStatus.COMPLETED},
     {},
     0, 0
    ),
    ({},
     {},
     {(1,0): JobStatus.COMPLETED},
     0, 1
    ),
    ({},
     {},
     {(1,0): JobStatus.FAILED},
     0, 1
    ),
])
async def test_Grid_check_queue_jel_mismatch(schedd, i3prod_path, set_time, jel_jobs, queue_jobs, hist_jobs, remove_calls, finish_calls):
    override = ['queue.type=htcondor', 'queue.max_task_queued_time=10', 'queue.max_task_processing_time=10', 'queue.suspend_submit_dir_time=10']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    qjobs = {}
    hjobs = {}
    g.submitter.get_jobs = MagicMock(return_value=qjobs)
    g.submitter.get_history = MagicMock()
    g.submitter.remove = MagicMock()
    g.finish = AsyncMock()
    g.get_tasks_on_queue = AsyncMock(return_value=[])

    jel = g.get_current_JEL()
    daydir = jel.parent
    def mkdir(name):
        p = daydir / name
        if p.exists():
            return p
        p.mkdir()
        return p

    for (c,p), s in jel_jobs.items():
        g.jobs[CondorJobId(cluster_id=c, proc_id=p)] = CondorJob(status=s, submit_dir=mkdir(f'{c}.{p}'))
    for (c,p), s in queue_jobs.items():
        qjobs[CondorJobId(cluster_id=c, proc_id=p)] = CondorJob(status=s, submit_dir=mkdir(f'{c}.{p}'))
    for (c,p), s in hist_jobs.items():
        hjobs[CondorJobId(cluster_id=c, proc_id=p)] = CondorJob(status=s, submit_dir=mkdir(f'{c}.{p}'))

    g.submitter.get_history.return_value = hjobs.items()
    await g.check()

    assert g.submitter.remove.call_count == remove_calls
    assert g.finish.call_count == finish_calls


@pytest.mark.parametrize('queue_jobs,hist_jobs,iceprod_tasks,reset_calls', [
    ({(1,0): ("dataset", "task", "instance")},
     {},
     [("dataset", "task", "instance", 0)],
     0,
    ),
    ({},
     {(1,0): ("dataset", "task", "instance")},
     [("dataset", "task", "instance", 1)],
     0,
    ),
    ({},
     {},
     [("dataset", "task", "instance", 1)],
     0,
    ),
    ({},
     {},
     [("dataset", "task", "instance", 600)],
     1,
    ),
])
async def test_Grid_check_queue_iceprod_mismatch(schedd, i3prod_path, set_time, queue_jobs, hist_jobs, iceprod_tasks, reset_calls):
    override = ['queue.type=htcondor', 'queue.max_task_queued_time=10', 'queue.max_task_processing_time=10', 'queue.suspend_submit_dir_time=10']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    qjobs = {}
    hjobs = {}
    g.submitter.get_jobs = MagicMock(return_value=qjobs)
    g.submitter.get_history = MagicMock()
    g.submitter.remove = MagicMock()
    g.finish = AsyncMock()
    g.task_reset = AsyncMock()
    itasks = []
    g.get_tasks_on_queue = AsyncMock(return_value=itasks)

    jel = g.get_current_JEL()
    daydir = jel.parent
    def mkdir(name):
        p = daydir / name
        if p.exists():
            return p
        p.mkdir()
        return p

    for (c,p), (d_id, t_id, i_id) in queue_jobs.items():
        qjobs[CondorJobId(cluster_id=c, proc_id=p)] = CondorJob(dataset_id=d_id, task_id=t_id, instance_id=i_id, submit_dir=mkdir(f'{c}.{p}'))
    for (c,p), (d_id, t_id, i_id) in hist_jobs.items():
        hjobs[CondorJobId(cluster_id=c, proc_id=p)] = CondorJob(dataset_id=d_id, task_id=t_id, instance_id=i_id, submit_dir=mkdir(f'{c}.{p}'))
    for (d_id, t_id, i_id, secs) in iceprod_tasks:
        t = set_time - datetime.timedelta(seconds=secs)
        itasks.append({'dataset_id': d_id, 'task_id': t_id, 'instance_id': i_id, 'status_changed': datetime2str(t)})

    g.submitter.get_history.return_value = hjobs.items()
    await g.check()

    assert g.task_reset.call_count == reset_calls


async def test_reset_task(schedd, i3prod_path, set_time):
    override = ['queue.type=htcondor', 'queue.max_task_queued_time=10', 'queue.max_task_processing_time=10', 'queue.suspend_submit_dir_time=10']
    cfg = iceprod.server.config.IceProdConfig(save=False, override=override)

    rc = MagicMock()
    g = iceprod.server.plugins.condor.Grid(cfg=cfg, rest_client=rc, cred_client=None)

    g.submitter.remove = MagicMock()

    jel = g.get_current_JEL()
    daydir = jel.parent
    p = daydir / 'olddir'
    p.mkdir()
    t = time.time() - 25
    os.utime(p, (t, t))
    logging.info('set time to %d', t)

    # normal failure
    jobid = CondorJobId(cluster_id=1, proc_id=0)
    g.jobs[jobid] = CondorJob(status=JobStatus.IDLE, submit_dir=p)

    g.task_success = AsyncMock()
    g.task_reset = AsyncMock()
    g.task_failure = AsyncMock()

    await g.finish(jobid, success=False)
    
    assert g.task_success.call_count == 0
    assert g.task_reset.call_count == 0
    assert g.task_failure.call_count == 1

    # success
    g.jobs[jobid] = CondorJob(status=JobStatus.IDLE, submit_dir=p)

    g.task_success = AsyncMock()
    g.task_reset = AsyncMock()
    g.task_failure = AsyncMock()

    await g.finish(jobid, success=True)
    
    assert g.task_success.call_count == 1
    assert g.task_reset.call_count == 0
    assert g.task_failure.call_count == 0

    # reset
    g.jobs[jobid] = CondorJob(status=JobStatus.IDLE, submit_dir=p)

    g.task_success = AsyncMock()
    g.task_reset = AsyncMock()
    g.task_failure = AsyncMock()

    await g.finish(jobid, success=False, reason=iceprod.server.plugins.condor.RESET_CONDOR_REASONS[0])

    assert g.task_success.call_count == 0
    assert g.task_reset.call_count == 1
    assert g.task_failure.call_count == 0

    # reset stderr
    g.jobs[jobid] = CondorJob(status=JobStatus.IDLE, submit_dir=p)

    g.task_success = AsyncMock()
    g.task_reset = AsyncMock()
    g.task_failure = AsyncMock()

    (p / 'condor.err').open('w').write(iceprod.server.plugins.condor.RESET_STDERR_REASONS[0])

    await g.finish(jobid, success=False)

    assert g.task_success.call_count == 0
    assert g.task_reset.call_count == 1
    assert g.task_failure.call_count == 0




DEFAULT_OPTIONS = {
    'dataset': 0,
    'job': 0,
    'jobs_submitted': 1,
    'resource_url': 'http://prod-exe.icecube.wisc.edu/',
    'data_url': 'gsiftp://gridftp.icecube.wisc.edu/',
    'svn_repository': 'http://code.icecube.wisc.edu/svn/',
    'site_temp': 'osdf:///icecube/wipac/scratch/simprod',
    'dataset_temp': '$(site_temp)/$(dataset)',
    'job_temp': '$(site_temp)/$(dataset)/$(job)',
}


def add_default_options(opts):
    """Add default config options"""
    for k in DEFAULT_OPTIONS:
        if k not in opts:
            opts[k] = DEFAULT_OPTIONS[k]



#!/usr/bin/env python

import os
import subprocess
import argparse
import json
import glob
import logging
import asyncio
from datetime import datetime

from rest_tools.client import RestClient

logger = logging.getLogger()

async def run(rest_client, dataset_id, job_start, job_stop, debug=True):
    async def update(job_id):
        logger.info('deleting job %s', job_id)
        await rest_client.request('PUT', f'/datasets/{dataset_id}/jobs/{job_id}/status', {'status':'suspended'})
        tasks = await rest_client.request('GET', f'/datasets/{dataset_id}/tasks', {'job_id':job_id, 'keys':'task_id'})
        for t in tasks:
            await rest_client.request('PUT', f'/datasets/{dataset_id}/tasks/{t}/status', {'status':'suspended'})
    try:
        jobs = await rest_client.request('GET', f'/datasets/{dataset_id}/jobs?keys=job_id|job_index')
        tasks = await rest_client.request('GET', f'/datasets/{dataset_id}/tasks?status=waiting&keys=task_id|job_id')
        for task_id in tasks:
            index = jobs[tasks[task_id]['job_id']]['job_index']
            if job_start <= index <= job_stop:
                await rest_client.request('PUT', f'/datasets/{dataset_id}/tasks/{task_id}/status', {'status':'queued'})

    except Exception:
        logger.error('error queueing tasks', exc_info=True)
        logger.info('%r',task_id)
        if debug:
            raise

def main():
    parser = argparse.ArgumentParser(description='clean up processing tasks not in pilots')
    parser.add_argument('-t', '--token',help='auth token')
    parser.add_argument('-d', '--dataset',help='dataset_id')
    parser.add_argument('job_start',type=int,help='job index start')
    parser.add_argument('job_stop',type=int,help='job index stop')
    
    args = parser.parse_args()
    args = vars(args)

    logging.basicConfig(level=logging.DEBUG)
    
    rpc = RestClient('https://iceprod2-api.icecube.wisc.edu', args['token'])

    asyncio.run(run(rpc, args['dataset'], args['job_start'], args['job_stop']))

if __name__ == '__main__':
    main()




#!/usr/bin/env python
import argparse
import logging
import asyncio

from rest_tools.client import DeviceGrantAuth, ClientCredentialsAuth


async def main():
    parser = argparse.ArgumentParser(description='job completion')

    parser.add_argument('--client-id-token', default='iceprod')
    parser.add_argument('client_secret_token')
    parser.add_argument('--client-id', default='iceprod')
    parser.add_argument('client_secret')
    parser.add_argument('--iceprod-url', default='https://credentials.iceprod.icecube.aq')
    parser.add_argument('--scratch-path', default='simprod')
    args = parser.parse_args()

    logging.basicConfig(level=logging.DEBUG)

    token_issuer = 'https://token-issuer.icecube.aq'

    token_kwargs = {
        'token_url': token_issuer,
        'client_id': args.client_id_token,
        'client_secret': args.client_secret_token,
        'scopes': [f'storage.read:/scratch/{args.scratch_path}', f'storage.modify:/scratch/{args.scratch_path}'],
    }
    rpc = DeviceGrantAuth('', **token_kwargs)
    rpc.refresh_token

    rpc_kwargs = {
        'token_url': 'https://keycloak.icecube.wisc.edu/auth/realms/IceCube',
        'client_id': args.client_id,
        'client_secret': args.client_secret,
    }
    rpc_cred = ClientCredentialsAuth(args.iceprod_url, **rpc_kwargs)

    try:
        await rpc_cred.request('POST', '/users/ice3simusr/credentials', {
            'url': token_issuer,
            'type': 'oauth',
            'transfer_prefix': f'osdf:///icecube/wipac/scratch/{args.scratch_path}',
            'refresh_token': rpc.refresh_token,
            'scope': ' '.join(token_kwargs['scopes'])
        })
    except Exception:
        await rpc_cred.request('PATCH', '/users/ice3simusr/credentials', {
            'url': token_issuer,
            'type': 'oauth',
            'transfer_prefix': f'osdf:///icecube/wipac/scratch/{args.scratch_path}',
            'refresh_token': rpc.refresh_token,
            'scope': ' '.join(token_kwargs['scopes'])
        })



if __name__ == '__main__':
    asyncio.run(main())



import logging
import json
import uuid

from iceprod.core.jsonUtil import json_encode
import tornado.web

from ..base_handler import APIBase
from ..auth import authorization, attr_auth
from iceprod.server.util import nowstr

logger = logging.getLogger('rest.task_stats')


def setup(handler_cfg):
    """
    Setup method for Task Stats REST API.

    Args:
        handler_cfg (dict): args to pass to the route

    Returns:
        dict: routes, database, indexes
    """
    return {
        'routes': [
            (r'/tasks/(?P<task_id>\w+)/task_stats', MultiTaskStatsHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/bulk/task_stats', DatasetsBulkTaskStatsHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/tasks/(?P<task_id>\w+)/task_stats', DatasetsMultiTaskStatsHandler, handler_cfg),
            (r'/datasets/(?P<dataset_id>\w+)/tasks/(?P<task_id>\w+)/task_stats/(?P<task_stat_id>\w+)', DatasetsTaskStatsHandler, handler_cfg),
        ],
        'database': 'task_stats',
        'indexes': {
            'jobs': {
                'task_stat_id_index': {'keys': 'task_stat_id', 'unique': True},
                'task_id_index': {'keys': 'task_id', 'unique': False},
                'dataset_id_index': {'keys': 'dataset_id', 'unique': False},
            }
        }
    }


class MultiTaskStatsHandler(APIBase):
    """
    Handle multi task_stats requests.
    """
    @authorization(roles=['admin', 'system'])
    async def post(self, task_id):
        """
        Create a task_stat entry.

        Body should contain the task stat data.

        Args:
            task_id (str): the task id for this task_stat

        Returns:
            dict: {'result': <task_stat_id>}
        """
        stat_data = json.loads(self.request.body)
        if 'dataset_id' not in stat_data:
            raise tornado.web.HTTPError(400, reason='Missing dataset_id in body')

        # set some fields
        task_stat_id = uuid.uuid1().hex
        data = {
            'task_stat_id': task_stat_id,
            'task_id': task_id,
            'dataset_id': stat_data.pop('dataset_id', ''),
            'create_date': nowstr(),
            'stats': stat_data,
        }

        await self.db.task_stats.insert_one(data)
        self.set_status(201)
        self.write({'result': task_stat_id})
        self.finish()


class DatasetsBulkTaskStatsHandler(APIBase):
    """
    Handle a dataset bulk task_stats requests.

    Stream the output of all stats.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='read')
    async def get(self, dataset_id):
        """
        Get task_stats for a dataset and task.

        Args:
            dataset_id (str): dataset id

        Params (optional):
            last: bool (True: only last task_stat.  False: all task_stats)
            after: only return stats created more recently than this date
            keys: | separated list of keys to return for each task_stat
            buffer_size: number of records to buffer before flushing (default 100)

        Returns:
            dict: {<task_stat_id>: stats}
        """
        last = self.get_argument('last', 'f').lower() in ('true','t','1','yes','y')

        query = {'dataset_id':dataset_id}
        after = self.get_argument('after', None)
        if after:
            query['create_date'] = {"$gte": after}

        projection = {'_id': False}
        keys = self.get_argument('keys','')
        if keys:
            keys = keys.split('|')
            projection.update({x:True for x in keys if x})
            projection['task_stat_id'] = True
            projection['task_id'] = True
            projection['create_date'] = True
        buffer_size = int(self.get_argument('buffer_size', '1000'))

        task_id = None
        data = []
        n = 0

        async def _write_data():
            nonlocal n
            ret = sorted(data, key=lambda x: x['create_date'])
            if keys:
                ret = [{k:d[k] for k in d if k in keys} for d in ret]
            if last:
                self.write(json_encode(ret[-1]) + '\n')
            else:
                for ret in data:
                    self.write(json_encode(ret) + '\n')
            n += 1
            if n >= buffer_size:
                n = 0
                await self.flush()

        async for row in self.db.task_stats.find(query, projection=projection).sort([('task_id',1)]):
            if row['task_id'] == task_id:
                data.append(row)
                continue
            if data:
                await _write_data()
            data = [row]
            task_id = row['task_id']

        if data:
            await _write_data()
        self.finish()


class DatasetsMultiTaskStatsHandler(APIBase):
    """
    Handle multi task_stats requests.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='read')
    async def get(self, dataset_id, task_id):
        """
        Get task_stats for a dataset and task.

        Args:
            dataset_id (str): dataset id
            task_id (str): task id

        Params (optional):
            last: bool (True: only last task_stat.  False: all task_stats)
            keys: | separated list of keys to return for each task_stat

        Returns:
            dict: {<task_stat_id>: stats}
        """
        last = self.get_argument('last', 'f').lower() in ('true','t','1','yes','y')

        projection = {'_id': False}
        keys = self.get_argument('keys','')
        if keys:
            projection.update({x:True for x in keys.split('|') if x})
            projection['task_stat_id'] = True
            if last:
                projection['create_date'] = True

        ret = await self.db.task_stats.find(
            {'dataset_id':dataset_id,'task_id':task_id},
            projection=projection
        ).to_list(10000)

        if last:
            ret = sorted(ret, key=lambda x: x['create_date'])[-1:]

        self.write({row['task_stat_id']:row for row in ret})
        self.finish()


class DatasetsTaskStatsHandler(APIBase):
    """
    Handle single task_stat requests.
    """
    @authorization(roles=['admin', 'user', 'system'])
    @attr_auth(arg='dataset_id', role='read')
    async def get(self, dataset_id, task_id, task_stat_id):
        """
        Get a task_stat entry.

        Args:
            dataset_id (str): dataset id
            task_id (str): task id
            task_stat_id (str): the task_stat id

        Returns:
            dict: task_stat entry
        """
        ret = await self.db.task_stats.find_one(
            {'dataset_id':dataset_id,'task_id':task_id,'task_stat_id':task_stat_id},
            projection={'_id':False}
        )
        if not ret:
            self.send_error(404, reason="Task stat not found")
        else:
            self.write(ret)
            self.finish()



from collections import Counter
import pytest
import requests.exceptions
from iceprod.server import states


async def test_rest_tasks_post(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']


async def test_rest_tasks_post_bad_role(server):
    client = server(roles=['user'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', '/tasks', data)
    assert exc_info.value.response.status_code == 403


async def test_rest_tasks_get(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    ret = await client.request('GET', '/tasks')
    assert 'tasks' in ret
    assert len(ret['tasks']) == 1
    assert ret['tasks'][0]['task_id'] == task_id

    ret = await client.request('GET', '/tasks', {'keys': 'dataset_id|name'})
    assert 'tasks' in ret
    assert len(ret['tasks']) == 1
    assert 'task_id' not in ret['tasks'][0]
    assert 'dataset_id' in ret['tasks'][0]
    assert 'name' in ret['tasks'][0]

    data = {'status': 'waiting'}
    await client.request('PUT', f'/tasks/{task_id}/status', data)

    ret = await client.request('GET', '/tasks', {'status': states.TASK_STATUS_START})
    assert 'tasks' in ret
    assert len(ret['tasks']) == 0

    ret = await client.request('GET', '/tasks', {'status': 'waiting'})
    assert 'tasks' in ret
    assert len(ret['tasks']) == 1
    assert ret['tasks'][0]['task_id'] == task_id


async def test_rest_tasks_get_by_site(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id1 = ret['result']
    await client.request('PATCH', f'/tasks/{task_id1}', {'site': 'foo'})

    ret = await client.request('POST', '/tasks', data)
    task_id2 = ret['result']
    await client.request('PATCH', f'/tasks/{task_id2}', {'site': 'bar'})

    ret = await client.request('POST', '/tasks', data)
    task_id3 = ret['result']
    await client.request('PATCH', f'/tasks/{task_id3}', {'site': 'foo.bar'})

    ret = await client.request('GET', '/tasks', {'site': 'foo'})
    assert 'tasks' in ret
    assert len(ret['tasks']) == 2
    assert [t['task_id'] for t in ret['tasks']] == [task_id1, task_id3]

    ret = await client.request('GET', '/tasks', {'site': 'bar'})
    assert 'tasks' in ret
    assert len(ret['tasks']) == 1
    assert [t['task_id'] for t in ret['tasks']] == [task_id2]

    ret = await client.request('GET', '/tasks', {'site': 'foo.bar'})
    assert 'tasks' in ret
    assert len(ret['tasks']) == 1
    assert [t['task_id'] for t in ret['tasks']] == [task_id3]


async def test_rest_tasks_get_details(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['task_id'] == task_id
    for k in data:
        assert k in ret
        assert data[k] == ret[k]
    for k in ('status','status_changed','failures','evictions','walltime',
              'walltime_err','walltime_err_n'):
        assert k in ret
    assert ret['status'] == states.TASK_STATUS_START


async def test_rest_tasks_patch(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    new_data = {
        'status': 'processing',
        'failures': 1,
    }
    ret = await client.request('PATCH', f'/tasks/{task_id}', new_data)
    for k in new_data:
        assert k in ret
        assert new_data[k] == ret[k]


async def test_rest_tasks_put_status(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
        'status': 'queued',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    new_data = {
        'status': 'failed',
    }
    await client.request('PUT', f'/tasks/{task_id}/status', new_data)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == new_data['status']


async def test_rest_tasks_dataset_get(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/tasks')
    assert task_id in ret
    for k in data:
        assert k in ret[task_id]
        assert ret[task_id][k] == data[k]


async def test_rest_tasks_dataset_get_details(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/tasks/{task_id}')
    for k in data:
        assert k in ret
        assert ret[k] == data[k]
    for k in ('status','status_changed','failures','evictions','walltime',
              'walltime_err','walltime_err_n'):
        assert k in ret
    assert ret['status'] == states.TASK_STATUS_START


async def test_rest_tasks_dataset_put_status(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
        'status': 'queued',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    new_data = {
        'status': 'failed',
    }
    await client.request('PUT', f'/datasets/{data["dataset_id"]}/tasks/{task_id}/status', new_data)

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/tasks/{task_id}')
    assert ret['status'] == new_data['status']


async def test_rest_tasks_dataset_summaries_status(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/task_summaries/status')
    assert ret == {states.TASK_STATUS_START: [task_id]}


async def test_rest_tasks_dataset_counts_status(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 1,
        'job_index': 0,
        'name': 'baz',
        'depends': [],
        'requirements': {'gpu': 1},
        'status': 'processing'
    }
    ret = await client.request('POST', '/tasks', data)

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/task_counts/status')
    assert ret == {states.TASK_STATUS_START: 1, 'processing': 1}

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/task_counts/status?status=complete')
    assert ret == {}
    
    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/task_counts/status?gpu=false')
    assert ret == {states.TASK_STATUS_START: 1}
    
    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/task_counts/status?gpu=true')
    assert ret == {'processing': 1}


async def test_rest_tasks_dataset_counts_name_status(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/task_counts/name_status')
    assert ret == {'bar': {states.TASK_STATUS_START: 1}}


async def test_rest_tasks_dataset_stats(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
        'status': 'processing',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/task_stats')
    assert ret == {}

    # mark complete to get a stat
    await client.request('PUT', f'/tasks/{task_id}/status', {'status': 'complete'})

    ret = await client.request('GET', f'/datasets/{data["dataset_id"]}/task_stats')
    assert 'bar' in ret
    for s in ('count','total_hrs','total_err_hrs','avg_hrs','stddev_hrs','min_hrs','max_hrs','efficiency'):
        assert s in ret['bar']


async def test_rest_tasks_actions_waiting(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'priority': .5,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    data = {
        'dataset_id': 'bar',
        'job_id': 'bar1',
        'task_index': 0,
        'job_index': 0,
        'priority': 10.,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id2 = ret['result']

    ret = await client.request('POST', '/task_actions/waiting', {'task_ids': [task_id2]})
    assert 'waiting' in ret
    assert ret['waiting'] == 1

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == states.TASK_STATUS_START

    ret = await client.request('GET', f'/tasks/{task_id2}')
    assert ret['status'] == 'waiting'


async def test_rest_tasks_actions_queue(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'status': 'waiting',
        'priority': .5,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    ret = await client.request('POST', '/task_actions/queue', {})
    assert ret['task_id'] == task_id
    assert ret['status'] == 'queued'

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'queued'


async def test_rest_tasks_actions_queue_reqs(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'status': 'waiting',
        'priority': .5,
        'name': 'bar',
        'depends': [],
        'requirements': {'memory': 4.5, 'disk': 100},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    # not enough reqs to queue task
    args = {'requirements': {'memory': 2.0, 'disk': 120}}
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', '/task_actions/queue', args)
    assert exc_info.value.response.status_code == 404

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'waiting'

    # now should queue
    args = {'requirements': {'memory': 6.0, 'disk': 120}}
    ret = await client.request('POST', '/task_actions/queue', args)
    assert task_id == ret['task_id']

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'queued'


@pytest.mark.parametrize('num_tasks,deprio,avail,out_tasks', [
    (10, [], {}, {}),
    (10, [], {'d1': 50}, {'d1': 10}),
    (10, ['d1'], {'d1': 5}, {'d1': 5}),
    (10, [], {'d1': 2, 'd2': 3}, {'d1': 2, 'd2': 3}),
    (10, ['d1', 'd2'], {'d1': 10, 'd2': 10, 'd3': 2}, {'d2': 8, 'd3': 2}),
    (10, ['d1', 'd2'], {'d1': 10, 'd2': 10, 'd3': 20}, {'d3': 10}),
    (10, ['d2', 'd1'], {'d1': 20, 'd2': 20}, {'d1': 10}),
    (10, ['d1', 'd2'], {'d1': 20, 'd2': 20}, {'d2': 10}),
])
async def test_rest_tasks_actions_queue_many(num_tasks, deprio, avail, out_tasks, server):
    client = server(roles=['system'])

    for d in avail:
        for n in range(avail[d]):
            data = {
                'dataset_id': d,
                'job_id': 'foo1',
                'task_index': 0,
                'job_index': n,
                'status': 'waiting',
                'priority': .5,
                'name': 'bar',
                'depends': [],
                'requirements': {},
            }
            task_id = await client.request('POST', '/tasks', data)
            ret = await client.request('GET', f'/tasks/{task_id["result"]}')
            assert ret['status'] == 'waiting'

    args = {
        'num': num_tasks,
        'dataset_deprio': deprio,
    }

    if not out_tasks:
        with pytest.raises(requests.exceptions.HTTPError) as exc_info:
            await client.request('POST', '/task_actions/queue_many', args)
        assert exc_info.value.response.status_code == 404
    else:
        ret = await client.request('POST', '/task_actions/queue_many', args)
        c = Counter(r['dataset_id'] for r in ret)
        assert c == out_tasks


async def test_rest_tasks_actions_process(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'status': 'queued',
        'priority': .5,
        'name': 'bar',
        'depends': [],
        'requirements': {},
        'instance_id': '12345',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    args = {
        'instance_id': '12345'
    }
    ret = await client.request('POST', f'/tasks/{task_id}/task_actions/processing', args)
    assert task_id == ret['task_id']

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'processing'
    assert ret['instance_id'] == args['instance_id']


async def test_rest_tasks_actions_reset(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'status': 'queued',
        'priority': .5,
        'name': 'bar',
        'depends': [],
        'requirements': {'memory':5.6, 'gpu':1},
        'instance_id': '12345',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    # try without instance id
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/tasks/{task_id}/task_actions/reset', {})
    assert exc_info.value.response.status_code == 400
    assert 'Missing instance_id' in exc_info.value.response.text

    # now with instance id
    args = {
        'instance_id': '12345',
    }
    await client.request('POST', f'/tasks/{task_id}/task_actions/reset', args)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'waiting'
    assert ret['instance_id'] == ''

    # now with bad instance id
    args = {
        'status': 'queued',
        'instance_id': '12345',
    }
    await client.request('PATCH', f'/tasks/{task_id}', args)

    args = {
        'instance_id': '666',
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/tasks/{task_id}/task_actions/reset', args)
    assert exc_info.value.response.status_code == 404

    # now try with time_used
    args = {
        'status': 'queued',
        'instance_id': '12345',
    }
    await client.request('PATCH', f'/tasks/{task_id}', args)

    # now try with time_used
    args = {
        'status': 'queued',
        'instance_id': '12345',
    }
    await client.request('PATCH', f'/tasks/{task_id}', args)

    args = {
        'instance_id': '12345',
        'time_used': 7200,
    }
    await client.request('POST', f'/tasks/{task_id}/task_actions/reset', args)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'waiting'
    assert ret['walltime_err_n'] == 1
    assert ret['walltime_err'] == 2.0

    # now try with resources
    args = {
        'status': 'queued',
        'instance_id': '12345',
    }
    await client.request('PATCH', f'/tasks/{task_id}', args)

    args = {
        'instance_id': '12345',
        'resources': {'time':2.5, 'memory':3.5, 'disk': 20.3, 'gpu': 23},
    }
    await client.request('POST', f'/tasks/{task_id}/task_actions/reset', args)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'waiting'
    assert ret['walltime_err_n'] == 2
    assert ret['walltime_err'] == 4.5
    assert ret['requirements']['memory'] == data['requirements']['memory']
    assert ret['requirements']['time'] >= args['resources']['time']
    assert ret['requirements']['disk'] >= args['resources']['disk']
    assert ret['requirements']['gpu'] != args['resources']['gpu']  # gpu doesn't change

    # now try with a bad status
    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 1,
        'job_index': 0,
        'status': 'complete',
        'priority': .5,
        'name': 'bar',
        'depends': [],
        'requirements': {'memory':5.6, 'gpu':1},
        'instance_id': '12345',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    args = {
        'instance_id': '12345',
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/tasks/{task_id}/task_actions/reset', args)
    assert exc_info.value.response.status_code == 400


async def test_rest_tasks_actions_failed(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'status': 'queued',
        'priority': .5,
        'name': 'bar',
        'depends': [],
        'requirements': {'memory':5.6, 'gpu':1},
        'instance_id': '12345',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    # try without instance id
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/tasks/{task_id}/task_actions/failed', {})
    assert exc_info.value.response.status_code == 400
    assert 'Missing instance_id' in exc_info.value.response.text

    # now with instance id
    args = {
        'instance_id': '12345',
    }
    await client.request('POST', f'/tasks/{task_id}/task_actions/failed', args)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'failed'

    # now with bad instance id
    args = {
        'status': 'queued',
        'instance_id': '12345',
    }
    await client.request('PATCH', f'/tasks/{task_id}', args)

    args = {
        'instance_id': '666',
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/tasks/{task_id}/task_actions/failed', args)
    assert exc_info.value.response.status_code == 404

    # now try with time_used
    args = {
        'status': 'queued',
        'instance_id': '12345',
    }
    await client.request('PATCH', f'/tasks/{task_id}', args)

    args = {
        'instance_id': '12345',
        'time_used': 7200
    }
    await client.request('POST', f'/tasks/{task_id}/task_actions/failed', args)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'failed'
    assert ret['walltime_err_n'] == 1
    assert ret['walltime_err'] == 2.0

    # now try with resources
    args = {
        'status': 'queued',
        'instance_id': '12345',
    }
    await client.request('PATCH', f'/tasks/{task_id}', args)

    args = {
        'instance_id': '12345',
        'resources': {'time':2.5, 'memory':3.5, 'disk': 20.3, 'gpu': 23}
    }
    await client.request('POST', f'/tasks/{task_id}/task_actions/failed', args)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'failed'
    assert ret['walltime_err_n'] == 2
    assert ret['walltime_err'] == 4.5
    assert ret['requirements']['memory'] == data['requirements']['memory']
    assert ret['requirements']['time'] >= args['resources']['time']
    assert ret['requirements']['disk'] >= args['resources']['disk']
    assert ret['requirements']['gpu'] != args['resources']['gpu']  # gpu doesn't change

    # now try with a bad status
    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 1,
        'job_index': 0,
        'status': 'complete',
        'priority': .5,
        'name': 'bar',
        'depends': [],
        'requirements': {'memory':5.6, 'gpu':1},
        'instance_id': '12345',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    args = {
        'instance_id': '12345',
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/tasks/{task_id}/task_actions/failed', args)
    assert exc_info.value.response.status_code == 400


async def test_rest_tasks_actions_complete(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'status': 'processing',
        'priority': .5,
        'name': 'bar',
        'depends': [],
        'requirements': {'memory':5.6, 'gpu':1},
        'instance_id': '12345',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    # try without instance id
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/tasks/{task_id}/task_actions/complete', {})
    assert exc_info.value.response.status_code == 400
    assert 'Missing instance_id' in exc_info.value.response.text

    # now with instance id
    args = {
        'instance_id': '12345',
    }
    await client.request('POST', f'/tasks/{task_id}/task_actions/complete', args)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'complete'

    # now with bad instance id
    args = {
        'status': 'processing',
        'instance_id': '12345',
    }
    await client.request('PATCH', f'/tasks/{task_id}', args)

    args = {
        'instance_id': '666',
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/tasks/{task_id}/task_actions/complete', args)
    assert exc_info.value.response.status_code == 404

    # now try with time_used
    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 1,
        'job_index': 0,
        'status': 'processing',
        'priority': .5,
        'name': 'bar',
        'depends': [],
        'requirements': {'memory':5.6, 'gpu':1},
        'instance_id': '12345',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    args = {
        'instance_id': '12345',
        'time_used': 7200
    }
    await client.request('POST', f'/tasks/{task_id}/task_actions/complete', args)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'complete'
    assert ret['walltime'] == 2.0

    # now try with a bad status
    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 2,
        'job_index': 0,
        'status': 'idle',
        'priority': .5,
        'name': 'bar',
        'depends': [],
        'requirements': {'memory':5.6, 'gpu':1},
        'instance_id': '12345',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    args = {
        'instance_id': '12345',
    }
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/tasks/{task_id}/task_actions/complete', args)
    assert exc_info.value.response.status_code == 400


async def test_rest_tasks_actions_bulk_status(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
        'status': 'queued',
        'instance_id': '12345',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    data2 = {'tasks': [task_id]}
    await client.request('POST', f'/datasets/{data["dataset_id"]}/task_actions/bulk_status/failed', data2)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'failed'
    assert ret['instance_id'] == ''

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo2',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id2 = ret['result']

    data2 = {'tasks': [task_id, task_id2]}
    await client.request('POST', f'/datasets/{data["dataset_id"]}/task_actions/bulk_status/waiting', data2)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'waiting'
    ret = await client.request('GET', f'/tasks/{task_id2}')
    assert ret['status'] == 'waiting'

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/datasets/{data["dataset_id"]}/task_actions/bulk_status/blah', data2)
    assert exc_info.value.response.status_code == 400

    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('POST', f'/datasets/{data["dataset_id"]}/task_actions/bulk_status/processing', {})
    assert exc_info.value.response.status_code == 400


async def test_rest_tasks_actions_bulk_suspend(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    data2 = {}
    await client.request('POST', f'/datasets/{data["dataset_id"]}/task_actions/bulk_suspend', data2)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'suspended'


async def test_rest_tasks_actions_bulk_suspend_by_job(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo2',
        'task_index': 0,
        'job_index': 1,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id2 = ret['result']

    data2 = {'jobs': ['foo1']}
    await client.request('POST', f'/datasets/{data["dataset_id"]}/task_actions/bulk_suspend', data2)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'suspended'
    ret = await client.request('GET', f'/tasks/{task_id2}')
    assert ret['status'] == states.TASK_STATUS_START


async def test_rest_tasks_actions_bulk_suspend_by_task(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo2',
        'task_index': 0,
        'job_index': 1,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id2 = ret['result']

    data2 = {'tasks': [task_id]}
    await client.request('POST', f'/datasets/{data["dataset_id"]}/task_actions/bulk_suspend', data2)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == 'suspended'
    ret = await client.request('GET', f'/tasks/{task_id2}')
    assert ret['status'] == states.TASK_STATUS_START


async def test_rest_tasks_actions_bulk_reset(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
        'status': 'waiting',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    data2 = {}
    await client.request('POST', f'/datasets/{data["dataset_id"]}/task_actions/bulk_reset', data2)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == states.TASK_STATUS_START


async def test_rest_tasks_actions_bulk_reset_by_job(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
        'status': 'waiting',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo2',
        'task_index': 0,
        'job_index': 1,
        'name': 'bar',
        'depends': [],
        'requirements': {},
        'status': 'waiting',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id2 = ret['result']

    data2 = {'jobs': ['foo1']}
    await client.request('POST', f'/datasets/{data["dataset_id"]}/task_actions/bulk_reset', data2)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == states.TASK_STATUS_START
    ret = await client.request('GET', f'/tasks/{task_id2}')
    assert ret['status'] == 'waiting'


async def test_rest_tasks_actions_bulk_reset_by_task(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
        'status': 'waiting',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo2',
        'task_index': 0,
        'job_index': 1,
        'name': 'bar',
        'depends': [],
        'requirements': {},
        'status': 'waiting',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id2 = ret['result']

    data2 = {'tasks': [task_id]}
    await client.request('POST', f'/datasets/{data["dataset_id"]}/task_actions/bulk_reset', data2)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == states.TASK_STATUS_START
    ret = await client.request('GET', f'/tasks/{task_id2}')
    assert ret['status'] == 'waiting'


async def test_rest_tasks_actions_bulk_hard_reset(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
        'status': 'complete',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    data2 = {}
    await client.request('POST', f'/datasets/{data["dataset_id"]}/task_actions/bulk_hard_reset', data2)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == states.TASK_STATUS_START


async def test_rest_tasks_actions_bulk_hard_reset_by_job(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
        'status': 'complete',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo2',
        'task_index': 0,
        'job_index': 1,
        'name': 'bar',
        'depends': [],
        'requirements': {},
        'status': 'complete',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id2 = ret['result']

    data2 = {'jobs': ['foo1']}
    await client.request('POST', f'/datasets/{data["dataset_id"]}/task_actions/bulk_hard_reset', data2)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == states.TASK_STATUS_START
    ret = await client.request('GET', f'/tasks/{task_id2}')
    assert ret['status'] == 'complete'


async def test_rest_tasks_actions_bulk_hard_reset_by_task(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
        'status': 'complete',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo2',
        'task_index': 0,
        'job_index': 1,
        'name': 'bar',
        'depends': [],
        'requirements': {},
        'status': 'complete',
    }
    ret = await client.request('POST', '/tasks', data)
    task_id2 = ret['result']

    data2 = {'tasks': [task_id]}
    await client.request('POST', f'/datasets/{data["dataset_id"]}/task_actions/bulk_hard_reset', data2)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['status'] == states.TASK_STATUS_START
    ret = await client.request('GET', f'/tasks/{task_id2}')
    assert ret['status'] == 'complete'


async def test_rest_tasks_actions_bulk_requirements(server):
    client = server(roles=['system'])

    data = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 0,
        'job_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    data2 = {
        'dataset_id': 'foo',
        'job_id': 'foo1',
        'task_index': 1,
        'job_index': 0,
        'name': 'baz',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data2)
    task_id2 = ret['result']

    args = {'cpu': 2}
    await client.request('PATCH', f'/datasets/{data["dataset_id"]}/task_actions/bulk_requirements/{data["name"]}', args)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert ret['requirements']['cpu'] == 2
    ret = await client.request('GET', f'/tasks/{task_id2}')
    assert 'cpu' not in ret['requirements']


    args = {'gpu': 4}
    await client.request('PATCH', f'/datasets/{data["dataset_id"]}/task_actions/bulk_requirements/{data2["name"]}', args)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert 'gpu' not in ret['requirements']
    ret = await client.request('GET', f'/tasks/{task_id2}')
    assert ret['requirements']['gpu'] == 4

    args = {'os': ['foo', 'bar']}
    await client.request('PATCH', f'/datasets/{data["dataset_id"]}/task_actions/bulk_requirements/{data2["name"]}', args)

    ret = await client.request('GET', f'/tasks/{task_id}')
    assert 'os' not in ret['requirements']
    ret = await client.request('GET', f'/tasks/{task_id2}')
    assert ret['requirements']['os'] == args['os']

    # test POST
    args = {'cpu': 1, 'memory': 2.5, 'os': []}
    await client.request('POST', f'/datasets/{data["dataset_id"]}/task_actions/bulk_requirements/{data2["name"]}', args)
    ret = await client.request('GET', f'/tasks/{task_id2}')
    assert ret['requirements'] == {'cpu': 1, 'memory': 2.5}

    # bad task name
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PATCH', f'/datasets/{data["dataset_id"]}/task_actions/bulk_requirements/blah', args)
    assert exc_info.value.response.status_code == 404

    # bad req
    args = {'blah': 4}
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PATCH', f'/datasets/{data["dataset_id"]}/task_actions/bulk_requirements/{data2["name"]}', args)
    assert exc_info.value.response.status_code == 400

    # bad req value
    args = {'memory': 'ten'}
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PATCH', f'/datasets/{data["dataset_id"]}/task_actions/bulk_requirements/{data2["name"]}', args)
    assert exc_info.value.response.status_code == 400

    # bad req value
    args = {'gpu': 3.5}
    with pytest.raises(requests.exceptions.HTTPError) as exc_info:
        await client.request('PATCH', f'/datasets/{data["dataset_id"]}/task_actions/bulk_requirements/{data2["name"]}', args)
    assert exc_info.value.response.status_code == 400


async def test_rest_tasks_files_get_empty(server):
    client = server(roles=['system'])

    dataset_id = 'foo'
    ret = await client.request('GET', f'/datasets/{dataset_id}/files')
    assert ret == {'files': []}


async def test_rest_tasks_files_post(server):
    client = server(roles=['system'])
    
    dataset_id = 'foo'
    data = {
        'dataset_id': dataset_id,
        'job_id': 'foo1',
        'job_index': 0,
        'task_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    file_data = {
        'job_index': data['job_index'],
        'task_name': data['name'],
        'filename': 'blah',
        'movement': 'input',
    }
    await client.request('POST', f'/datasets/{dataset_id}/files', file_data)

    ret = await client.request('GET', f'/datasets/{dataset_id}/files')
    assert len(ret['files']) == 1
    assert ret['files'][0]['remote'] == file_data['filename']


async def test_rest_tasks_files_task_get(server):
    client = server(roles=['system'])
    
    dataset_id = 'foo'
    data = {
        'dataset_id': dataset_id,
        'job_id': 'foo1',
        'job_index': 0,
        'task_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    file_data = {
        'job_index': data['job_index'],
        'task_name': data['name'],
        'filename': 'blah',
        'movement': 'input',
    }
    await client.request('POST', f'/datasets/{dataset_id}/files', file_data)

    ret = await client.request('GET', f'/datasets/{dataset_id}/files/{task_id}')
    assert len(ret['files']) == 1
    assert ret['files'][0]['remote'] == file_data['filename']


async def test_rest_tasks_files_task_post(server):
    client = server(roles=['system'])
    
    dataset_id = 'foo'
    data = {
        'dataset_id': dataset_id,
        'job_id': 'foo1',
        'job_index': 0,
        'task_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    file_data = {
        'filename': 'blah',
        'movement': 'input',
    }
    await client.request('POST', f'/datasets/{dataset_id}/files/{task_id}', file_data)

    ret = await client.request('GET', f'/datasets/{dataset_id}/files/{task_id}')
    assert len(ret['files']) == 1
    assert ret['files'][0]['remote'] == file_data['filename']


async def test_rest_tasks_files_task_delete(server):
    client = server(roles=['system'])
    
    dataset_id = 'foo'
    data = {
        'dataset_id': dataset_id,
        'job_id': 'foo1',
        'job_index': 0,
        'task_index': 0,
        'name': 'bar',
        'depends': [],
        'requirements': {},
    }
    ret = await client.request('POST', '/tasks', data)
    task_id = ret['result']

    file_data = {
        'filename': 'blah',
        'movement': 'input',
    }
    await client.request('POST', f'/datasets/{dataset_id}/files/{task_id}', file_data)
    await client.request('DELETE', f'/datasets/{dataset_id}/files/{task_id}')

    ret = await client.request('GET', f'/datasets/{dataset_id}/files/{task_id}')
    assert len(ret['files']) == 0



import asyncio

from .server import Server
from ..core.logger import stderr_logger


stderr_logger()


async def main():
    s = Server()
    await s.start()
    try:
        await asyncio.Event().wait()
    finally:
        await s.stop()


asyncio.run(main())



"""
Test script for util
"""

from __future__ import absolute_import, division, print_function

from tests.util import unittest_reporter, glob_tests

import logging
logger = logging.getLogger('util')


try:
    pass
except:
    pass

import unittest
import iceprod.core.util


class util_test(unittest.TestCase):
    def setUp(self):
        super(util_test,self).setUp()

    def tearDown(self):
        super(util_test,self).tearDown()

    @unittest_reporter
    def test_01_NoncriticalError(self):
        """Test the NoncriticalError class"""
        e = iceprod.core.util.NoncriticalError()

        if not isinstance(e,Exception):
            raise Exception('NoncriticalError is not an Exception')
        if e.value != '':
            raise Exception('Empty NoncriticalError has non-empty value')
        if str(e) != "NoncriticalError()":
            logger.info('repr: %s',str(e))
            raise Exception('bad string representation of NoncriticalError')

        e = iceprod.core.util.NoncriticalError('some text')

        if not isinstance(e,Exception):
            raise Exception('NoncriticalError is not an Exception')
        if e.value != 'some text':
            raise Exception('NoncriticalError has different value')
        if str(e) != "NoncriticalError('some text')":
            logger.info('repr: %s',str(e))
            raise Exception('bad string representation of NoncriticalError')

def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    alltests = glob_tests(loader.getTestCaseNames(util_test))
    suite.addTests(loader.loadTestsFromNames(alltests,util_test))
    return suite



"""
Some Prometheus utilities.
"""
import asyncio
import time

from prometheus_client import Histogram
from tornado.web import RequestHandler
from wipac_dev_tools.prometheus_tools import GlobalLabels, AsyncPromWrapper


class HistogramBuckets:
    """Prometheus histogram buckets"""

    # DEFAULT = [.005, .01, .025, .05, .075, .1, .25, .5, .75, 1, 2.5, 5, 7.5, 10]

    #: Database bucket centered around 5ms, with outliers up to 10s
    DB = [.001, .0025, .005, .0075, .01, .025, .05, .1, .25, .5, 1, 10]

    #: API bucket centered around 50ms, up to 10s
    API = [.005, .01, .025, .04, .05, .06, .075, .1, .25, .5, 1, 5, 10]

    #: Timer bucket up to 1 second
    SECOND = [.0001, .0005, .001, .0025, .005, .0075, .01, .025, .05, .075, .1, .25, .5, .75, 1]

    #: Timer bucket up to 10 seconds
    TENSECOND = [.001, .0025, .005, .0075, .01, .025, .05, .075, .1, .25, .5, .75, 1, 2.5, 5, 10]

    #: Timer bucket up to 1 minute
    MINUTE = [.1, .5, 1, 2.5, 5, 7.5, 10, 15, 20, 25, 30, 45, 60]

    #: Timer bucket up to 10 minutes
    TENMINUTE = [1, 5, 10, 15, 20, 25, 30, 45, 60, 90, 120, 150, 180, 240, 300, 360, 420, 480, 540, 600]

    #: Timer bucket up to 1 hour
    HOUR = [10, 60, 120, 300, 600, 1200, 1800, 2400, 3000, 3600]


class PromRequestMixin(RequestHandler):
    PromHTTPHistogram = Histogram('http_request_duration_seconds', 'HTTP request duration in seconds', labelnames=('method', 'handler', 'status'), buckets=HistogramBuckets.API)

    def prepare(self):
        super().prepare()
        self._prom_start_time = time.monotonic()

    def on_finish(self):
        super().on_finish()
        end_time = time.monotonic()
        self.PromHTTPHistogram.labels(
            method=str(self.request.method).lower(),
            handler=f'{self.__class__.__module__.split(".")[-1]}.{self.__class__.__name__}',
            status=str(self.get_status()),
        ).observe(end_time - self._prom_start_time)


class AsyncMonitor(GlobalLabels):
    SLEEP_TIME = 5

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._task = None

    async def start(self):
        if self._task is None:
            self._task = asyncio.create_task(self._monitor())

    async def stop(self):
        if self._task:
            self._task.cancel()
            try:
                await self._task
            except asyncio.CancelledError:
                pass
            finally:
                self._task = None

    @AsyncPromWrapper(lambda self: self.prometheus.gauge('asyncio_tasks_running', 'Python asyncio tasks active'))
    async def _monitor(self, prom_gauge):
        while True:
            prom_gauge.set(len(asyncio.all_tasks()))
            await asyncio.sleep(self.SLEEP_TIME)



import os
import pytest


@pytest.fixture(scope='module')
def mongo_url(monkeymodule):
    if 'DB_URL' not in os.environ:
        monkeymodule.setenv('DB_URL', 'mongodb://localhost/creds')


@pytest.fixture(autouse=True)
def rest_dbs(monkeypatch):
    monkeypatch.setenv('DATABASES', 'creds')



"""
Test script for jsonUtil
"""

from __future__ import absolute_import, division, print_function

from tests.util import unittest_reporter, glob_tests

import logging
logger = logging.getLogger('jsonUtil')

import os
import time
import shutil

try:
    pass
except:
    pass

import unittest

from datetime import date,datetime,time
import iceprod.core.jsonUtil


class jsonUtil_test(unittest.TestCase):
    def setUp(self):
        super(jsonUtil_test,self).setUp()
        self.test_dir = os.path.join(os.getcwd(),'test')
        if not os.path.exists(self.test_dir):
            os.mkdir(self.test_dir)

    def tearDown(self):
        shutil.rmtree(self.test_dir)
        super(jsonUtil_test,self).tearDown()

    @unittest_reporter
    def test_01_recursive_unicode(self):
        """Test recursive_unicode"""
        # test byte string
        input = 'a test'
        expected = u'a test'
        output = iceprod.core.jsonUtil.recursive_unicode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test tuple of strings
        input = ('a test','another test')
        expected = (u'a test',u'another test')
        output = iceprod.core.jsonUtil.recursive_unicode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test list of strings
        input = ['a test','another test']
        expected = [u'a test',u'another test']
        output = iceprod.core.jsonUtil.recursive_unicode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test dict of strings
        input = {'a':'a test','b':'another test'}
        expected = {u'a':u'a test',u'b':u'another test'}
        output = iceprod.core.jsonUtil.recursive_unicode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test something that is not a string
        input = 1
        expected = 1
        output = iceprod.core.jsonUtil.recursive_unicode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

    @unittest_reporter
    def test_02_objToJSON(self):
        """Test objToJSON"""
        # test byte string
        input = 'a test'
        expected = 'a test'
        output = iceprod.core.jsonUtil.objToJSON(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test int
        input = 1
        expected = 1
        output = iceprod.core.jsonUtil.objToJSON(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test dict
        input = {'a':1}
        expected = {'a':1}
        output = iceprod.core.jsonUtil.objToJSON(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test datetime
        input = datetime(2012,3,6,12,34,29)
        expected = {'__jsonclass__': ['datetime', '2012-03-06T12:34:29']}
        output = iceprod.core.jsonUtil.objToJSON(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test datetime
        input = datetime(2012,3,6,12,34,29,120)
        expected = {'__jsonclass__': ['datetime', '2012-03-06T12:34:29.000120']}
        output = iceprod.core.jsonUtil.objToJSON(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test date
        input = date(2012,3,6)
        expected = {'__jsonclass__': ['date', '2012-03-06']}
        output = iceprod.core.jsonUtil.objToJSON(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test time
        input = time(12,34,29)
        expected = {'__jsonclass__': ['time', '12:34:29']}
        output = iceprod.core.jsonUtil.objToJSON(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test time
        input = time(12,34,29,120)
        expected = {'__jsonclass__': ['time', '12:34:29.000120']}
        output = iceprod.core.jsonUtil.objToJSON(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

    @unittest_reporter
    def test_03_JSONToObj(self):
        """Test JSONToObj"""
        # test byte string
        input = 'a test'
        expected = 'a test'
        output = iceprod.core.jsonUtil.JSONToObj(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test int
        input = 1
        expected = 1
        output = iceprod.core.jsonUtil.JSONToObj(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test dict
        input = {'a':1}
        expected = {'a':1}
        output = iceprod.core.jsonUtil.JSONToObj(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test datetime
        input = {'__jsonclass__': ['datetime', '2012-03-06T12:34:29']}
        expected = datetime(2012,3,6,12,34,29)
        output = iceprod.core.jsonUtil.JSONToObj(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test datetime
        input = {'__jsonclass__': ['datetime', '2012-03-06T12:34:29.000120']}
        expected = datetime(2012,3,6,12,34,29,120)
        output = iceprod.core.jsonUtil.JSONToObj(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test datetime
        input = {'__jsonclass__': ['datetime', '2012-03-06 12:34:29']}
        expected = datetime(2012,3,6,12,34,29)
        output = iceprod.core.jsonUtil.JSONToObj(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test date
        input = {'__jsonclass__': ['date', '2012-03-06']}
        expected = date(2012,3,6)
        output = iceprod.core.jsonUtil.JSONToObj(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test time
        input = {'__jsonclass__': ['time', '12:34:29']}
        expected = time(12,34,29)
        output = iceprod.core.jsonUtil.JSONToObj(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test time
        input = {'__jsonclass__': ['time', '12:34:29.000120']}
        expected = time(12,34,29,120)
        output = iceprod.core.jsonUtil.JSONToObj(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test binary
        input = {"__jsonclass__":["binary","dGVzdGluZw=="]}
        expected = 'testing'
        output = iceprod.core.jsonUtil.JSONToObj(input)
        self.assertEqual(expected, output)

    @unittest_reporter
    def test_05_json_encode(self):
        """Test json_encode"""
        # test byte string
        input = 'a test'
        expected = '"a test"'
        output = iceprod.core.jsonUtil.json_encode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test tuple of strings
        input = ('a test','another test')
        expected = '["a test","another test"]'
        output = iceprod.core.jsonUtil.json_encode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test list of strings
        input = ['a test','another test']
        expected = '["a test","another test"]'
        output = iceprod.core.jsonUtil.json_encode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test dict of strings
        input = {'a':'a test','b':'another test'}
        expected = '{"a":"a test","b":"another test"}'
        output = iceprod.core.jsonUtil.json_encode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test int
        input = 1
        expected = '1'
        output = iceprod.core.jsonUtil.json_encode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test float
        input = 1.05
        expected = '1.05'
        output = iceprod.core.jsonUtil.json_encode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test datetime
        input = datetime(2012,3,6,12,34,29)
        expected = '{"__jsonclass__":["datetime","2012-03-06T12:34:29"]}'
        output = iceprod.core.jsonUtil.json_encode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test datetime
        input = datetime(2012,3,6,12,34,29,120)
        expected = '{"__jsonclass__":["datetime","2012-03-06T12:34:29.000120"]}'
        output = iceprod.core.jsonUtil.json_encode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test date
        input = date(2012,3,6)
        expected = '{"__jsonclass__":["date","2012-03-06"]}'
        output = iceprod.core.jsonUtil.json_encode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test time
        input = time(12,34,29)
        expected = '{"__jsonclass__":["time","12:34:29"]}'
        output = iceprod.core.jsonUtil.json_encode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test time
        input = time(12,34,29,120)
        expected = '{"__jsonclass__":["time","12:34:29.000120"]}'
        output = iceprod.core.jsonUtil.json_encode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

    @unittest_reporter
    def test_06_json_decode(self):
        """Test json_decode"""
        # test byte string
        input = '"a test"'
        expected = 'a test'
        output = iceprod.core.jsonUtil.json_decode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test list of strings
        input = '["a test","another test"]'
        expected = ['a test','another test']
        output = iceprod.core.jsonUtil.json_decode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test dict of strings
        input = '{"a":"a test","b":"another test"}'
        expected = {'a':'a test','b':'another test'}
        output = iceprod.core.jsonUtil.json_decode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test int
        input = '1'
        expected = 1
        output = iceprod.core.jsonUtil.json_decode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test float
        input = '1.05'
        expected = 1.05
        output = iceprod.core.jsonUtil.json_decode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test datetime
        input = '{"__jsonclass__":["datetime","2012-03-06T12:34:29"]}'
        expected = datetime(2012,3,6,12,34,29)
        output = iceprod.core.jsonUtil.json_decode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test datetime
        input = '{"__jsonclass__":["datetime","2012-03-06T12:34:29.000120"]}'
        expected = datetime(2012,3,6,12,34,29,120)
        output = iceprod.core.jsonUtil.json_decode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test date
        input = '{"__jsonclass__":["date","2012-03-06"]}'
        expected = date(2012,3,6)
        output = iceprod.core.jsonUtil.json_decode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test time
        input = '{"__jsonclass__":["time","12:34:29"]}'
        expected = time(12,34,29)
        output = iceprod.core.jsonUtil.json_decode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test time
        input = '{"__jsonclass__":["time","12:34:29.000120"]}'
        expected = time(12,34,29,120)
        output = iceprod.core.jsonUtil.json_decode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))

        # test binary
        input = '{"__jsonclass__":["binary","dGVzdGluZw=="]}'
        expected = 'testing'
        output = iceprod.core.jsonUtil.json_decode(input)
        if expected != output:
            raise Exception('expected != output:  %r != %r'%(expected,output))


def load_tests(loader, tests, pattern):
    suite = unittest.TestSuite()
    alltests = glob_tests(loader.getTestCaseNames(jsonUtil_test))
    suite.addTests(loader.loadTestsFromNames(alltests,jsonUtil_test))
    return suite


"""
Main website
------------

This is the external website users will see when interacting with IceProd.
It has been broken down into several sub-handlers for easier maintenance.
"""

import importlib.resources
import logging
import os
import random
import re

from prometheus_client import Info, start_http_server
import tornado.web
from rest_tools.client import RestClient, ClientCredentialsAuth
from rest_tools.server import catch_error, RestServer, RestHandlerSetup, OpenIDLoginHandler
from rest_tools.server.session import Session

from iceprod.util import VERSION_STRING
from iceprod.common.prom_utils import AsyncMonitor, PromRequestMixin
from iceprod.core.config import ConfigSchema as DATASET_SCHEMA
from iceprod.server.config import CONFIG_SCHEMA as SERVER_SCHEMA
from iceprod.server import documentation
from iceprod.server.util import nowstr

from .config import get_config
from .handlers.base import authenticated, LoginMixin, PublicHandler
from .handlers.submit import Config, ConfigStatus, Submit, SubmitStatus
from .handlers.dataset import Dataset, DatasetBrowse
from .handlers.job import Job, JobBrowse
from .handlers.task import Task, TaskBrowse

logger = logging.getLogger('website')


class Default(PublicHandler):
    """Handle / urls"""
    @catch_error
    async def get(self):
        # try to get the user, if available
        await self.get_current_user_async()
        self.render('main.html')


class Schemas(PublicHandler):
    """Handle /schemas/v3/(.*) urls"""
    @catch_error
    async def get(self, schema):
        if schema == 'dataset.schema.json':
            self.write(DATASET_SCHEMA.schema())
        elif schema == 'config.schema.json':
            self.write(SERVER_SCHEMA)
        elif ver := re.match(r'dataset_v(\d\.\d).schema.json', schema):
            ver = float(ver.group(1))
            self.write(DATASET_SCHEMA.schema(ver))
        else:
            raise tornado.web.HTTPError(404, reason='unknown schema')


class Documentation(PublicHandler):
    @catch_error
    async def get(self, url):
        # try to get the user, if available
        await self.get_current_user_async()
        doc_path = str(importlib.resources.files('iceprod.server')/'data'/'docs')
        full_path = os.path.join(doc_path, url)
        if not full_path.startswith(doc_path):
            self.set_status(404)
            self.render('404.html', path='bad docs path')
            return
        self.write(documentation.load_doc(full_path))
        self.flush()


class Log(PublicHandler):
    @authenticated
    async def get(self, dataset_id, log_id):
        assert self.rest_client
        ret = await self.rest_client.request('GET','/datasets/{}/logs/{}'.format(dataset_id, log_id))
        log_text = ret['data']
        html = '<html><head><title>' + ret['name'] + '</title></head><body>'
        html += log_text.replace('\n', '<br/>')
        html += '</body></html>'
        self.write(html)
        self.flush()


class Help(PublicHandler):
    """Help Page"""
    @catch_error
    async def get(self):
        # try to get the user, if available
        await self.get_current_user_async()
        self.render('help.html')


class Other(PublicHandler):
    """Handle any other urls - this is basically all 404"""
    @catch_error
    async def get(self):
        # try to get the user, if available
        await self.get_current_user_async()
        path = self.request.path
        self.set_status(404)
        self.render('404.html', path=path)


class Profile(PublicHandler):
    """Handle user profile page"""
    @authenticated
    async def get(self):
        username = self.current_user
        groups = self.auth_groups
        group_creds = {}
        for g in groups:
            if g != 'users':
                group_creds[g] = await self.get_cred_group_tokens(group_name=g)
        user_creds = await self.get_cred_user_tokens(username=username)
        self.render('profile.html', username=username, groups=groups,
                    group_creds=group_creds, user_creds=user_creds)


class Login(LoginMixin, PromRequestMixin, OpenIDLoginHandler):  # type: ignore
    pass


class Logout(PublicHandler):
    @catch_error
    async def get(self):
        self.clear_tokens()
        self.current_user = None
        self.request.uri = '/'  # for login redirect, fake the main page
        self.render('logout.html', status=None)


class HealthHandler(PublicHandler):
    """
    Handle health requests.
    """
    async def get(self):
        """
        Get health status.

        Returns based on exit code, 200 = ok, 400 = failure
        """
        status = {
            'now': nowstr(),
        }

        try:
            await self.system_rest_client.request('GET', '/dataset_summaries/status')
            status['rest_api'] = 'OK'
        except Exception:
            logger.info('error from REST API', exc_info=True)
            self.send_error(500, reason='error from REST API')
            return

        try:
            await self.cred_rest_client.request('GET', '/healthz')
            status['cred_service'] = 'OK'
        except Exception:
            logger.info('error from REST API', exc_info=True)
            self.send_error(500, reason='error from credential service')
            return

        self.write(status)


class Server:
    def __init__(self):
        config = get_config()

        # get package data
        static_path = str(importlib.resources.files('iceprod.website')/'data'/'www')
        if static_path is None or not os.path.exists(static_path):
            logger.info('static path: %r',static_path)
            raise Exception('bad static path')
        template_path = str(importlib.resources.files('iceprod.website')/'data'/'www_templates')
        if template_path is None or not os.path.exists(template_path):
            logger.info('template path: %r',template_path)
            raise Exception('bad template path')

        # set IceProd REST API
        if config.ICEPROD_API_ADDRESS:
            rest_address = config.ICEPROD_API_ADDRESS
        else:
            raise RuntimeError('ICEPROD_API_ADDRESS not specified')

        rest_config = {
            'debug': config.DEBUG,
            'server_header': 'IceProd/' + VERSION_STRING,
        }

        if config.OPENID_URL:
            logging.info(f'enabling auth via {config.OPENID_URL} for aud "{config.OPENID_AUDIENCE}"')
            rest_config.update({
                'auth': {
                    'openid_url': config.OPENID_URL,
                    'audience': config.OPENID_AUDIENCE,
                }
            })
        elif config.CI_TESTING:
            rest_config.update({
                'auth': {
                    'secret': 'secret',
                }
            })
        else:
            raise RuntimeError('OPENID_URL not specified, and CI_TESTING not enabled!')

        # enable monitoring
        self.prometheus_port = config.PROMETHEUS_PORT if config.PROMETHEUS_PORT > 0 else None
        self.async_monitor = None

        if config.ICEPROD_CRED_CLIENT_ID and config.ICEPROD_CRED_CLIENT_SECRET:
            logging.info(f'enabling auth via {config.OPENID_URL} for aud "{config.OPENID_AUDIENCE}"')
            cred_client = ClientCredentialsAuth(
                address=config.ICEPROD_CRED_ADDRESS,
                token_url=config.OPENID_URL,
                client_id=config.ICEPROD_CRED_CLIENT_ID,
                client_secret=config.ICEPROD_CRED_CLIENT_SECRET,
            )
        elif config.CI_TESTING:
            cred_client = RestClient(config.ICEPROD_CRED_ADDRESS, timeout=1, retries=0)
        else:
            raise RuntimeError('ICEPROD_CRED_CLIENT_ID or ICEPROD_CRED_CLIENT_SECRET not specified, and CI_TESTING not enabled!')

        handler_args = RestHandlerSetup(rest_config)
        if config.CI_TESTING:
            self.session = Session()
        else:
            kwargs = {}
            if config.REDIS_USER:
                kwargs['username'] = config.REDIS_USER
            if config.REDIS_PASSWORD:
                kwargs['password'] = config.REDIS_PASSWORD
            if config.REDIS_TLS:
                kwargs['tls'] = True
            self.session = Session(storage_type='redis', host=config.REDIS_HOST, **kwargs)  # type: ignore
        handler_args['session'] = self.session

        full_url = config.ICEPROD_WEB_URL
        login_url = full_url+'/login'

        login_handler_args = handler_args.copy()
        if config.ICEPROD_API_CLIENT_ID and config.ICEPROD_API_CLIENT_SECRET:
            logging.info('enabling system rest client and website logins"')
            rest_client = ClientCredentialsAuth(
                address=config.ICEPROD_API_ADDRESS,
                token_url=config.OPENID_URL,
                client_id=config.ICEPROD_API_CLIENT_ID,
                client_secret=config.ICEPROD_API_CLIENT_SECRET,
            )
            login_handler_args['oauth_client_id'] = config.ICEPROD_API_CLIENT_ID
            login_handler_args['oauth_client_secret'] = config.ICEPROD_API_CLIENT_SECRET
            login_handler_args['oauth_client_scope'] = 'offline_access posix profile'
        elif config.CI_TESTING:
            logger.info('CI_TESTING: no login for testing')
            rest_client = RestClient(config.ICEPROD_API_ADDRESS, timeout=1, retries=0)
        else:
            raise RuntimeError('ICEPROD_API_CLIENT_ID or ICEPROD_API_CLIENT_SECRET not specified, and CI_TESTING not enabled!')

        handler_args.update({
            'rest_api': rest_address,
            'cred_rest_client': cred_client,
            'system_rest_client': rest_client,
            'auth_url': config.OPENID_URL,
            'auth_client_id': config.ICEPROD_API_CLIENT_ID,
            'auth_client_secret': config.ICEPROD_API_CLIENT_SECRET,
        })
        if config.COOKIE_SECRET:
            cookie_secret = config.COOKIE_SECRET
            log_cookie_secret = cookie_secret[:4] + 'X'*(len(cookie_secret)-8) + cookie_secret[-4:]
            logger.info('using supplied cookie secret %r', log_cookie_secret)
        else:
            cookie_secret = ''.join(hex(random.randint(0,15))[-1] for _ in range(64))

        server = RestServer(
            debug=config.DEBUG,
            cookie_secret=cookie_secret,
            login_url=login_url,
            template_path=template_path,
            static_path=static_path,
        )

        server.add_route("/", Default, handler_args)
        server.add_route('/config', Config, handler_args)
        server.add_route(r'/config/status/(\w+)', ConfigStatus, handler_args)
        server.add_route(r"/schemas/v3/([\w\.]+)", Schemas, handler_args)
        server.add_route('/dataset', DatasetBrowse, handler_args)
        server.add_route(r"/dataset/(\w+)", Dataset, handler_args)
        server.add_route(r"/dataset/(\w+)/task", TaskBrowse, handler_args)
        server.add_route(r"/dataset/(\w+)/task/(\w+)", Task, handler_args)
        server.add_route(r"/dataset/(\w+)/job", JobBrowse, handler_args)
        server.add_route(r"/dataset/(\w+)/job/(\w+)", Job, handler_args)
        server.add_route('/help', Help, handler_args)
        server.add_route(r"/docs/(.*)", Documentation, handler_args)
        server.add_route(r"/dataset/(\w+)/log/(\w+)", Log, handler_args)
        server.add_route('/submit', Submit, handler_args)
        server.add_route(r'/submit/status/(\w+)', SubmitStatus, handler_args)
        server.add_route('/profile', Profile, handler_args)
        server.add_route('/login', Login, login_handler_args)
        server.add_route('/logout', Logout, handler_args)

        server.add_route('/healthz', HealthHandler, handler_args)
        server.add_route(r"/.*", Other, handler_args)

        server.startup(address=config.HOST, port=config.PORT)

        self.server = server

    async def start(self):
        if self.prometheus_port:
            logging.info("starting prometheus on %r", self.prometheus_port)
            start_http_server(self.prometheus_port)
            i = Info('iceprod', 'IceProd information')
            i.info({
                'version': VERSION_STRING,
                'type': 'website',
            })
            self.async_monitor = AsyncMonitor(labels={'type': 'website'})
            await self.async_monitor.start()

    async def stop(self):
        await self.server.stop()
        if self.async_monitor:
            await self.async_monitor.stop()
        self.session.close()



"""
Manage resources like CPUs and Memory.  Default values, detection, tracking,
policy, etc.
"""

from collections import OrderedDict
import logging
import math
import os


# default rounding bins for resources
RESOURCE_BINS = {
    'cpu': list(range(1, 1000)),
    'gpu': list(range(0, 100)),
    'memory': [x/10. for x in list(range(5, 50, 5)) + list(range(50, 200, 10)) + list(range(200, 1000, 40)) + list(range(1000, 10000, 100))],
    'disk': list(range(1, 10)) + list(range(10, 50, 2)) + list(range(50, 100, 5)) + list(range(100, 2000, 20)),
    'time': [x/60. for x in list(range(10, 60, 10)) + list(range(60, 360, 15)) + list(range(360, 1440, 60)) + list(range(1440, 20240, 240))],
}


class Resources:
    """
    Task (and node) resource definition and tracking.
    """

    # ordering based on priority for matching
    defaults = OrderedDict([
        ('gpu', []),
        ('memory', 2.0),
        ('disk', 10.0),
        ('time', 8.0),
        ('cpu', 1),
    ])
    """Default resource values

    cpu
        integer

    gpu
        fungible - a list of unique items

    memory
        float, in GB

    disk
        float, in GB

    time
        float, in hours
    """

    overusage_limits = {
        'cpu': {'ignore': 2., 'allowed': 4., 'movingavg':10},
        'gpu': {'ignore': 1.5, 'allowed': 1.5, 'movingavg':10},
        'memory': {'ignore': 0, 'allowed': 10.},
        'disk': {'ignore': 0, 'allowed': 10.},
        'time': {'ignore': 0, 'allowed': 10.},
    }
    """Overusage limits

    ignore
        Ignore extra usage up to this ratio over the allocated resource.

    allowed
        Allow extra usage if there are available resources, up to this ratio.

    movingavg
        Track a moving average of this many measurements, to avoid spikes.
    """


def du(path):
    """
    Perform a "du" on a path, getting the disk usage.

    Args:
        path (str): The path to analyze

    Returns:
        int: bytes used
    """
    logging.info('du of %s', path)
    total = 0
    for root,dirs,files in os.walk(path):
        for d in list(dirs):
            p = os.path.join(root,d)
            if os.path.islink(p):
                dirs.remove(d)
        for f in files:
            p = os.path.join(root,f)
            if not os.path.islink(p):
                total += os.path.getsize(p)
    logging.info('du of %s finished: %r', path, total)
    return total


def group_hasher(resources):
    """
    Hash a set of resources into a binned group.
    """
    ret = 0
    if 'cpu' in resources:
        ret = int(resources['cpu'])
    if 'gpu' in resources:
        if isinstance(resources['gpu'],(int,float)):
            ret ^= int(resources['gpu'])*100
        else:
            ret ^= len(resources['gpu'])*100
    if 'memory' in resources:
        ret ^= int(math.log(resources['memory'])*math.e)*1000
    if 'disk' in resources:
        ret ^= int(math.log(resources['disk'])*math.e)*1000000
    if 'time' in resources:
        ret ^= int(resources['time']) * 1000000000
    if 'os' in resources:
        ret ^= hash(resources['os']) & (0b11111111 << 32)
    return ret


def sanitized_requirements(reqs, use_defaults=False):
    """
    Sanitize a dict of requirements (resources) for a task.

    Args:
        reqs (dict): dict of requirements
    Returns:
        dict: sanitized requirements
    """
    reqs = {k.lower():v for k,v in reqs.items()}
    ret = {}
    all_keys = set(reqs).union(Resources.defaults)
    for k in all_keys:
        if k in reqs:
            try:
                if k in Resources.defaults:
                    if isinstance(Resources.defaults[k], (int,list)):
                        ret[k] = int(reqs[k])
                    elif isinstance(Resources.defaults[k], float):
                        ret[k] = float(reqs[k])
                else:
                    ret[k] = reqs[k]
            except ValueError:
                pass
        if use_defaults and k not in ret and k in Resources.defaults:
            if isinstance(Resources.defaults[k], (int,float,str)):
                ret[k] = Resources.defaults[k]
    return ret


def rounded_requirements(reqs, bins=None):
    """
    Round requirements into bins for submit systems to have
    fewer "choices" to consider for job matching.

    Args:
        reqs (dict): dict of requirements
        bins (dict): dict of binnings to use (None for defaults)
    Returns:
        dict: rounded requirements
    """
    if not bins:
        bins = RESOURCE_BINS

    def round_up(num, bins):
        """Round up to the next bin value"""
        for b in bins:
            if num <= b*1.05:  # within 5%
                return b
        raise Exception('num too big for bin sizes')
    ret = {}
    for k in reqs:
        v = reqs[k]
        if k in bins:
            v = round_up(v, bins[k])
        ret[k] = v
    return ret



"""
IceProd Config Migration:
- remove all "resources"
- remove all "classes"
- remove all "running_class"
  - use env shell to track down source?
- data compression removed
- remove all "steering/system"

IceProd Dataset Migration:
- dataset status truncated moved to attrinute
  - status -> suspended
- update all tasks ("waiting", "queued", "processing", "reset") -> "idle"
"""

import argparse
from getpass import getpass
import json
import jsonschema
import logging
from pathlib import Path
from pprint import pprint

from pymongo import MongoClient
from iceprod.core.config import Dataset


notbool = lambda x: not x
strcommalist = lambda x: ','.join(x)


def convert(config):
    logging.info('Now converting config')

    if (steering := config.get('steering', {})) and steering.get('system', None):
        del config['steering']['system']

    def data_cleaner(obj):
        ret = []
        if 'batchsys' in obj:
            if not obj['batchsys']:
                obj['batchsys'] = {}
        if 'resources' in obj:
            if obj['resources']:
                for r in obj['resources']:
                    r['type'] = 'permanent'
                    r['movement'] = 'input'
                    if 'compression' in r:
                        del r['compression']
                    ret.append(r)
            del obj['resources']
        if 'data' in obj:
            if (data := obj['data']) and isinstance(data, list):
                mydata = []
                for d in data:
                    if d.get('type', 'permanent') not in ('permanent', 'job_temp', 'dataset_temp', 'site_temp'):
                        logging.warning('%r', d)
                        raise Exception('unknown data type')
                    if 'compression' in d:
                        del d['compression']
                    if d.get('transfer') == 'exists':
                        d['transfer'] = 'maybe'
                    add = True
                    if d.get('type') == 'job_temp' and d.get('movement') == 'input':
                        for dd in list(mydata):
                            if dd.get('type') == 'job_temp' and dd.get('movement') == 'output' and d.get('local') == dd.get('local'):
                                mydata.remove(dd)
                                add = False
                                break
                    if add:
                        mydata.append(d)
                obj['data'] = mydata
                ret.extend(mydata)
            #del obj['data']
        if 'classes' in obj:
            for cl in obj['classes']:
                if cl.get('libs'):
                    raise Exception('classes with libs not allowed')
                ret.append({
                    'type': 'permanent',
                    'movement': 'input',
                    'remote': cl['src'],
                })
            del obj['classes']
        if 'system' in obj:
            del obj['system']
        return ret

    if steering := config.get('steering', {}):
        base_data = data_cleaner(steering)
    else:
        base_data = []
        config['steering'] = {}

    for i,task in enumerate(config.get('tasks', [])):
        if 'name' not in task:
            task['name'] = f'Task{i}'
        data = base_data.copy()
        data.extend(data_cleaner(task))
        for tray in task.get('trays', []):
            data_cleaner(tray)
            for module in tray.get('modules', []):
                data_cleaner(module)
        task['data'] = data

    config['version'] = 3.1
    return config


def conversion(config, output=None):
    logging.info('Converting config')
    new_config = convert(config)
    logging.info('Validating config')
    d = Dataset('', 1, 1, 1, 1, 'processing', 1., 'group', 'user', False, new_config)
    d.fill_defaults()
    d.validate()
    logging.info('Config validated!')

    if output == '-':
        print(json.dumps(d.config, indent=2, sort_keys=True))
    elif output:
        with open(output, 'w') as f:
            json.dump(d.config, f, indent=2, sort_keys=True)
    return d.config


def do_mongo(server, dataset_id=None, output=None, dryrun=False):
    username = input('MongoDB Username:')
    password = getpass('MongoDB Password:')

    port = 27017
    if ':' in server:
        server,port = server.split(':')
        port = int(port)
    if server.startswith('mongodb://'):
        server = server[10:]
    client = MongoClient(server, port, username=username, password=password)
    db = client.config

    datasets = {}
    search = {}
    if dataset_id:
        datasets[dataset_id] = {'dataset': 'specifed'}
        search = {'dataset_id': dataset_id}
    else:
        for d in client.datasets.datasets.find({}, projection={"_id": False, "dataset_id": True, "dataset": True, "status": True}):
            if d['dataset'] > 22000 or d['status'] == 'processing':
                datasets[d['dataset_id']] = d

    for config in db.config.find(search, projection={'_id': False}):
        if config['dataset_id'] not in datasets:
            continue
        dataset = datasets[config['dataset_id']]
        try:
            logging.warning('Processing %r', dataset['dataset'])
            new_config = conversion(config, output=output)
        except jsonschema.exceptions.ValidationError:
            if dataset['status'] != 'processing' and not config['tasks'][0].get('trays'):
                logging.info('skipping dataset %r due to errors', dataset['dataset'])
                continue
            print(json.dumps(config, indent=2, sort_keys=True))
            raise
        except Exception:
            print(json.dumps(config, indent=2, sort_keys=True))
            raise
        assert config['dataset_id'] == new_config['dataset_id']
        if not dryrun:
            db.config.find_one_and_replace({'dataset_id': config['dataset_id']}, new_config)
        logging.warning('Completed %r', dataset['dataset'])


def main():
    parser = argparse.ArgumentParser()
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-c', '--config', help='config json file')
    group.add_argument('--mongo-server', default=None, help='mongodb server address')
    parser.add_argument('-o', '--output', default=None, help='output config json to file (or "-" for stdout)')
    parser.add_argument('--dataset-id', default=None, help='dataset id (for mongo)')
    parser.add_argument('--dry-run', action='store_true', default=False, help='do a dry run (for mongo)')
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)

    if args.config:
        logging.info('Opening config at %s', args.config)
        with open(args.config) as f:
            config = json.load(f)
        conversion(config, output=args.output)
    elif args.mongo_server:
        do_mongo(args.mongo_server, dataset_id=args.dataset_id, output=args.output, dryrun=args.dry_run)

if __name__ == '__main__':
    main()


"""
Test script for scheduled_tasks/job_completion
"""

import logging
from unittest.mock import MagicMock

import pytest
from iceprod.scheduled_tasks import job_completion

logger = logging.getLogger('scheduled_tasks_job_completion_test')


async def test_200_run():
    rc = MagicMock()
    dataset_summaries = {}
    job_summaries = {}
    tasks = {
        't1':{'task_id':'t1','task_index':1,'status':'processing'},
        't2':{'task_id':'t2','task_index':2,'status':'waiting'},
    }
    async def client(method, url, args=None):
        logger.info('REST: %s, %s', method, url)
        if url.startswith('/dataset_summaries'):
            return dataset_summaries
        if url.startswith('/datasets/foo/job_summaries'):
            return job_summaries
        elif url.startswith('/datasets/foo/tasks'):
            return tasks
        elif url == '/datasets/foo/jobs/1/status' and method == 'PUT':
            client.called = True
            client.status = args['status']
            return {}
        else:
            raise Exception()
    client.called = False
    client.status = None
    rc.request = client

    await job_completion.run(rc, debug=True)
    assert not client.called

    dataset_summaries['processing'] = ['foo']
    await job_completion.run(rc, debug=True)
    assert not client.called

    job_summaries['processing'] = ['1']
    await job_completion.run(rc, debug=True)
    assert not client.called

    logger.info('test processing')
    await job_completion.run(rc, debug=True)
    assert not client.called

    logger.info('test errors')
    tasks['t1']['status'] = 'failed'
    tasks['t2']['status'] = 'failed'
    await job_completion.run(rc, debug=True)
    assert client.called
    assert client.status == 'errors'

    logger.info('test processing and suspended')
    tasks['t1']['status'] = 'processing'
    tasks['t2']['status'] = 'suspended'
    client.called = False
    client.status = None
    await job_completion.run(rc, debug=True)
    assert not client.called

    logger.info('test suspended')
    tasks['t1']['status'] = 'complete'
    tasks['t2']['status'] = 'suspended'
    await job_completion.run(rc, debug=True)
    assert client.called
    assert client.status == 'suspended'

    logger.info('test complete')
    tasks['t1']['status'] = 'complete'
    tasks['t2']['status'] = 'complete'
    client.called = False
    client.status = None
    await job_completion.run(rc, debug=True)
    assert client.called
    assert client.status == 'complete'


async def test_201_run():
    rc = MagicMock()
    async def client(method, url, args=None):
        logger.info('REST: %s, %s', method, url)
        if url.startswith('/dataset_summaries'):
            return {'processing':['foo']}
        else:
            raise Exception()
    rc.request = client
    with pytest.raises(Exception):
        await job_completion.run(rc, debug=True)

    # check it normally hides the error
    await job_completion.run(rc, debug=False)
